"""TcEx Framework Module"""
# standard library
from typing import Literal

# third-party
from rich import print  # pylint: disable=redefined-builtin
from rich.panel import Panel
from rich.prompt import Prompt

AlignMethod = Literal['left', 'center', 'right']


class RenderPrompt:
    """Render Prompt"""

    title_align: AlignMethod = 'left'

    @staticmethod
    def ask(
        text: str,
        choices: list[str] | None = None,
        default: str | None = None,
        password: bool = False,
        show_choices: bool = True,
        show_default: bool = True,
    ) -> str | None:
        """Render a prompt"""
        return Prompt.ask(
            text,
            choices=choices,
            default=default,
            password=password,
            show_choices=show_choices,
            show_default=show_default,
        )

    @classmethod
    def input(cls, prompt_text: str, prompt_default: str, subtitle: str = '') -> str:
        """Render a prompt"""
        prompt_text = f'[white]{prompt_text}[/white][bold white]{prompt_default}[/bold white]'
        print(Panel(prompt_text, title='Input', title_align=cls.title_align, subtitle=subtitle))

        # collect input
        return input('> ').strip()  # nosec



"""TcEx Framework Module"""

# first-party
from tcex.input.model.advanced_request_model import AdvancedRequestModel
from tcex.input.model.aot_execution_enabled_model import AotExecutionEnabledModel
from tcex.input.model.app_api_service_model import AppApiServiceModel
from tcex.input.model.app_external_model import AppExternalModel
from tcex.input.model.app_feed_api_service_model import AppFeedApiServiceModel
from tcex.input.model.app_organization_model import AppOrganizationModel
from tcex.input.model.app_playbook_model import AppPlaybookModel
from tcex.input.model.app_system_model import AppSystemModel
from tcex.input.model.app_trigger_service_model import AppTriggerServiceModel
from tcex.input.model.app_webhook_trigger_service_model import AppWebhookTriggerServiceModel
from tcex.input.model.cal_setting_model import CalSettingModel

# from tcex.input.model.service_model import ServiceModel
from tcex.input.model.smtp_setting_model import SmtpSettingModel

# define feature to model map
feature_map = {
    'aotExecutionEnabled': [AotExecutionEnabledModel],
    'appBuilderCompliant': [],
    # 'advancedRequest': [],
    'CALSettings': [CalSettingModel],
    'fileParams': [],
    'layoutEnabledApp': [],
    'secureParams': [],
    'smtpSettings': [SmtpSettingModel],
    # features for TC Playbook loop prevention
    'CreatesGroup': [],
    'CreatesIndicator': [],
    'CreatesSecurityLabel': [],
    'CreatesTag': [],
    'DeletesGroup': [],
    'DeletesIndicator': [],
    'DeletesSecurityLabel': [],
    'DeletesTag': [],
}

# define runtime level to model map
runtime_level_map = {
    'apiservice': AppApiServiceModel,
    'feedapiservice': AppFeedApiServiceModel,
    'external': AppExternalModel,
    'organization': AppOrganizationModel,
    'playbook': AppPlaybookModel,
    # special case for non-supported system based Apps
    'system': AppSystemModel,
    'triggerservice': AppTriggerServiceModel,
    'webhooktriggerservice': AppWebhookTriggerServiceModel,
}

tc_action_map = {
    'Advanced Request': [AdvancedRequestModel],
}



"""TcEx Framework Module"""
# third-party
from requests import Session

# first-party
from tcex.api.tc.v3.case_attributes.case_attribute import CaseAttribute, CaseAttributes


class CaseManagement:
    """Case Management

    Args:
        session: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, session: Session):
        """Initialize instance properties."""
        self.session = session

    def artifact(self, **kwargs) -> 'Artifact':
        """Return a instance of Artifact object.

        Model Schema:
        >>> from tcex.api.tc.v3.case_management.model.artifact_model import ArtifactModel
        >>> print(ArtifactModel.schema_json(by_alias=False, indent=2))

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            case_id (int, kwargs): The **case id** for the Artifact.
            case_xid (str, kwargs): The **case xid** for the Artifact.
            derived_link (bool, kwargs): Flag to specify if this artifact should be used for
                potentially associated cases or not.
            file_data (str, kwargs): Base64 encoded file attachment required only for certain
                artifact types.
            hash_code (str, kwargs): Hashcode of Artifact of type File.
            notes (Note, kwargs): A list of Notes corresponding to the Artifact.
            source (str, kwargs): The **source** for the Artifact.
            summary (str, kwargs): The **summary** for the Artifact.
            task_id (int, kwargs): The ID of the task which the Artifact references.
            task_xid (str, kwargs): The XID of the task which the Artifact references.
            type (str, kwargs): The **type** for the Artifact.
        """
        return Artifact(session=self.session, **kwargs)

    def artifacts(self, **kwargs) -> 'Artifacts':
        """Return a instance of Artifacts object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Artifact.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return Artifacts(session=self.session, **kwargs)

    def artifact_type(self, **kwargs) -> 'ArtifactType':
        """Return a instance of Artifact Type object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            data_type (str, kwargs): [Read-Only] The **Data Type** for the Artifact Type.
            description (str, kwargs): [Read-Only] The **Description** for the Artifact Type.
            intel_type (str, kwargs): [Read-Only] The **Intel Type** for the Artifact Type.
            name (str, kwargs): [Read-Only] The **Name** for the Artifact Type.
        """
        return ArtifactType(session=self.session, **kwargs)

    def artifact_types(self, **kwargs) -> 'ArtifactTypes':
        """Return a instance of Artifact Types object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent
                while retrieving the Artifact Type objects.
        """
        return ArtifactTypes(session=self.session, **kwargs)

    def case(self, **kwargs) -> 'Case':
        """Return a instance of Case object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            artifacts (Artifacts, kwargs): A list of Artifacts corresponding to the Case.
            assignee (Assignee, kwargs): The user or group Assignee object for the Case.
            description (str, kwargs): The **description** for the Case.
            name (str, kwargs): The **name** for the Case.
            notes (Notes, kwargs): A list of Notes corresponding to the Case.
            resolution (str, kwargs): The **resolution** for the Case.
            severity (str, kwargs): The **severity** for the Case.
            status (str, kwargs): The **status** for the Case.
            tags (Notes, kwargs): A list of Tags corresponding to the Case (NOTE: Setting this
                parameter will replace any existing tag(s) with the one(s) specified).
            tasks (Tasks, kwargs): A list of Tasks corresponding to the Case.
            user_access (Users, kwargs): A list of Users that, when defined, are the only ones
                allowed to view or edit the Case.
            workflow_events (WorkflowEvents, kwargs): A list of workflowEvents (timeline)
                corresponding to the Case.
            workflow_template (WorkflowTemplate, kwargs): The Template that the Case is populated
                by.
            xid (str, kwargs): The **xid** for the Case.
        """
        return Case(session=self.session, **kwargs)

    def case_attribute(self, **kwargs) -> CaseAttribute:
        """Return a instance of Case Attributes object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            case_id (int, kwargs): Case associated with attribute.
            default (bool, kwargs): A flag indicating that this is the default attribute of its type
                within the object. Only applies to certain attribute and data types.
            source (str, kwargs): The attribute source.
            type (str, kwargs): The attribute type.
            value (str, kwargs): Attribute value.
        """
        return CaseAttribute(session=self.session, **kwargs)

    def case_attributes(self, **kwargs) -> CaseAttributes:
        """Return a instance of Case Attributes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Group.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return CaseAttributes(session=self.session, **kwargs)

    def cases(self, **kwargs) -> 'Cases':
        """Return a instance of Cases object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Case objects.
        """
        return Cases(session=self.session, **kwargs)

    def create_entity(self, entity: dict, owner: str) -> dict | None:
        """Create a CM object provided a dict and owner."""
        entity_type = entity.pop('type').lower()
        entity_type = entity_type.replace(' ', '_')
        try:
            obj = getattr(self, entity_type)(**entity)
        except AttributeError:
            return None

        obj.create()
        data = {'status_code': obj.request.status_code}
        if obj.request.ok:
            data.update(obj.request.json().get('data', {}))
            data['main_type'] = 'Case_Management'
            data['sub_type'] = entity_type
            data['owner'] = owner

        return data

    def note(self, **kwargs) -> 'Note':
        """Return a instance of Note object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifact (Artifact, kwargs): [Read-Only] The **Artifact** for the Note.
            artifact_id (int, kwargs): the ID of the Artifact on which to apply the Note
            author (int, kwargs): [Read-Only] The **Author** for the Note.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Note.
            case_xid (str, kwargs): [Required (alt: caseId)]
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Note.
            edited (boolean, kwargs): [Read-Only] The **Edited** for the Note.
            last_modified (str, kwargs): [Read-Only] The **Last Modified** for the Note.
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Note.
            summary (str, kwargs): [Read-Only] The **Summary** for the Note.
            task (Task, kwargs): [Read-Only] The **Task** for the Note.
            task_id (int, kwargs): the ID of the Task on which to apply the Note
            task_xid (str, kwargs): the XID of the Task on which to apply the Note
            text (str, kwargs): [Required] The **Text** for the Note.
            workflow_event (WorkflowEvent, kwargs): [Read-Only] The **Workflow Event** for the Note.
            workflow_event_id (int, kwargs): the ID of the Event on which to apply the Note
        """
        return Note(session=self.session, **kwargs)

    def notes(self, **kwargs) -> 'Notes':
        """Return a instance of Notes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Note.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Notes objects.
        """
        return Notes(session=self.session, **kwargs)

    def tag(self, **kwargs) -> 'Tag':
        """Return a instance of Tag object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Tag.
            description (str, kwargs): a brief description of the Tag
            last_used (str, kwargs): [Read-Only] The **Last Used** for the Tag.
            name (str, kwargs): [Required] The **Name** for the Tag.
            owner (str, kwargs): [Read-Only] The **Owner** for the Tag.
        """
        return Tag(session=self.session, **kwargs)

    def tags(self, **kwargs) -> 'Tags':
        """Return a instance of Tags object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Tag.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Tag objects.
        """
        return Tags(session=self.session, **kwargs)

    def task(self, **kwargs) -> 'Task':
        """Return a instance of Task object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Task
            assignee (Assignee, kwargs): the user or group Assignee object for the Task
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Task.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Task.
            completed_by (str, kwargs): [Read-Only] The **Completed By** for the Task.
            completed_date (str, kwargs): the completion date of the Task
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Task.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Task.
            dependent_on_id (str, kwargs): [Read-Only] The **Dependent On ID** for the Task.
            description (str, kwargs): The **Description** for the Task.
            due_date (str, kwargs): the due date of the Task
            duration (int, kwargs): [Read-Only] The **Duration** for the Task.
            id_dependent_on (int, kwargs): [Read-Only] The **Id_Dependent On** for the Task.
            name (str, kwargs): [Required] The **Name** for the Task.
            notes (Note, kwargs): a list of Notes corresponding to the Task
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Task.
            required (bool, kwargs): [Read-Only] The **Required** flag for the Task.
            status (str, kwargs): The **Status** for the Task.
            workflow_phase (int, kwargs): the phase of the workflow
            workflow_step (int, kwargs): the step of the workflow
            xid (str, kwargs): The **Xid** for the Task.
        """
        return Task(session=self.session, **kwargs)

    def tasks(self, **kwargs) -> 'Tasks':
        """Return a instance of Tasks object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Task.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Task objects.
        """
        return Tasks(session=self.session, **kwargs)

    def workflow_event(self, **kwargs) -> 'WorkflowEvent':
        """Return a instance of Workflow Event object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Workflow Event.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid**
                for the Workflow Event.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Workflow Event.
            deleted (bool, kwargs): [Read-Only] The **Deleted** flag for the Workflow Event.
            deleted_reason (str, kwargs): the reason for deleting the event (required
                input for DELETE operation only)
            event_date (str, kwargs): the time that the Event is logged
            link (str, kwargs): [Read-Only] The **Link** for the Workflow Event.
            link_text (str, kwargs): [Read-Only] The **Link Text** for the Workflow Event.
            notes (Note, kwargs): a list of Notes corresponding to the Event
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Workflow Event.
            summary (str, kwargs): [Required] The **Summary** for the Workflow Event.
            system_generated (bool, kwargs): [Read-Only] The **System Generated**
                flag for the Workflow Event.
            user (User, kwargs): [Read-Only] The **User** for the Workflow Event.
        """
        return WorkflowEvent(session=self.session, **kwargs)

    def workflow_events(self, **kwargs) -> 'WorkflowEvents':
        """Return a instance of Workflow Events object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowEvents(session=self.session, **kwargs)

    def workflow_template(self, **kwargs) -> 'WorkflowTemplate':
        """Return a instance of Workflow Template object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): [Read-Only] The **Active** flag for the Workflow Template.
            assignee (Assignee, kwargs): [Read-Only] The **Assignee** for the Workflow Template.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Workflow Template.
            config_artifact (str, kwargs): [Read-Only] The **Config Artifact** for the Workflow
                Template.
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Workflow
                Template.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Workflow Template.
            description (str, kwargs): The **Description** for the Workflow Template.
            name (str, kwargs): [Required] The **Name** for the Workflow Template.
            target_type (str, kwargs): [Read-Only] The **Target Type** for the Workflow Template.
            version (int, kwargs): The **Version** for the Workflow Template.
        """
        return WorkflowTemplate(session=self.session, **kwargs)

    def workflow_templates(self, **kwargs) -> 'WorkflowTemplates':
        """Return a instance of Workflow Templates object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Workflow Template.
            tql_filters (list, optional): List of TQL filters.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowTemplates(session=self.session, **kwargs)


# first-party
# pylint: disable=wrong-import-position
from tcex.api.tc.v3.artifact_types.artifact_type import ArtifactType, ArtifactTypes
from tcex.api.tc.v3.artifacts.artifact import Artifact, Artifacts
from tcex.api.tc.v3.cases.case import Case, Cases
from tcex.api.tc.v3.notes.note import Note, Notes
from tcex.api.tc.v3.tags.tag import Tag, Tags
from tcex.api.tc.v3.tasks.task import Task, Tasks
from tcex.api.tc.v3.workflow_events.workflow_event import WorkflowEvent, WorkflowEvents
from tcex.api.tc.v3.workflow_templates.workflow_template import WorkflowTemplate, WorkflowTemplates



"""TcEx Framework Module"""
# standard library
import functools
from collections.abc import Callable, Container
from typing import TYPE_CHECKING, Any, TypeVar

# third-party
from redis import Redis

if TYPE_CHECKING:
    from .app.app import App
    from .app.playbook.playbook import Playbook
    from .app.token import Token
    from .exit.exit import Exit
    from .input.input import Input
    from .logger.logger import Logger
    from .requests_tc import RequestsTc, TcSession

T = TypeVar('T')


class Registry(Container):
    """Dynamic service registry that supports raw values, factories, and factory providers.

    Terms:
        Service - An object that is registered with this registry and can be retrieved from it by
            type or name.
        Factory - A callable that can return an instance that fulfills a type or name.
        Provider - An object with one or more members that are factories (as denoted by the
            @factory decorator)
    """

    def __init__(self):
        """Initialize instance properties."""
        self._values = {}

    def add_service(self, type_or_name: str | type, value: Any):
        """Add an instance of a type to this registry.

        A service is a single instance of the given type.

        Args:
            type_or_name: the concrete type of the provided service, or a name (MyClass or
            'MyClass')
            value: The service.
        """
        self._add(type_or_name, value)

    def add_method(self, method: Callable):
        """Add an instance method to the registry.

        Args:
            method: A instance method to add to the registry.
        """
        self._add(method.__name__, method)

    def add_factory(self, type_or_name: str | type, factory: Callable, singleton=False):
        """Add a factory for a service.

        A factory is any callable that can be invoked to provide the given type of service.

        Args:
            type_or_name: the concrete type of the provided service, or a name (MyClass or
            'MyClass')
            factory: the callable that can create the above type.
            singleton: if True, the factory will be invoked exactly once.  If False, the factory
            will be invoked every time the service is requested.
        """
        self._add(type_or_name, (singleton, factory))

    def register(self, provider):
        """Add a provider, which provides one or more service factories.

        A provider is a class with one or more members decorated with Registry@factory.
        When you add a provider with this method, each of its decorated members will be added to
        this registry as a factory.
        """
        for entry in dir(provider):
            try:
                provider_function = type(provider).__dict__[entry]
                factory_provider = getattr(provider_function, 'factory_provider', None)
                if factory_provider:
                    provided_type, singleton = factory_provider
                    if callable(provider_function):  # A function or member function
                        # if it's a bound method, this will get the bound version
                        provider_member = getattr(provider, entry)
                        self.add_factory(provided_type, provider_member, singleton)
                    elif hasattr(provider_function, '__get__'):
                        # this is a property or non-callable descriptor:
                        self.add_factory(
                            provided_type,
                            functools.partial(provider_function.__get__, provider, provider),
                            singleton,
                        )
                    else:
                        self.add_service(provided_type, provider_function)
            except KeyError:
                pass

    def _add(self, type_or_name: str | type, value: Callable | type | tuple[bool, Callable]):
        """Add a service to the registry."""
        key = type_or_name if isinstance(type_or_name, str) else type_or_name.__name__
        self._values[key] = value

    def __getattr__(self, type_or_name):
        """Enable property-access style access to registered services, i.e., registry.MyClass."""
        return self._retrieve_registered_value(type_or_name)

    def __contains__(self, item: Any) -> bool:
        """Enable the syntax MyClass in registry."""
        try:
            self.__getattr__(item)
            return True
        except RuntimeError:
            return False

    def _retrieve_registered_value(self, type_or_name):
        """Retrieve or create a value, if already registered."""
        key = type_or_name if isinstance(type_or_name, str) else type_or_name.__name__
        if key in self._values:
            value = self._values[key]
            if isinstance(value, tuple):
                singleton, factory = value
                if not singleton:
                    return factory()

                value = factory()
                self._values[key] = value

            return value

        raise RuntimeError(f'No provider for type: {key}')

    def _reset(self):
        """Only used during testing to reset registry."""
        self._values = {}

    @staticmethod
    def factory(type_or_name: str | type, singleton: bool = False) -> Callable[[T], T]:
        """Decorate a function that can be treated as a factory that provides a service.

        Args:
            type_or_name: the concrete type of the provided service, or a name (MyClass or
            'MyClass')
            singleton: if True, the factory will be invoked exactly once.  If False, the factory
            will be invoked every time the service is requested.
        """

        def _decorator(original: T) -> T:
            setattr(original, 'factory_provider', (type_or_name, singleton))
            return original

        return _decorator

    #
    # The below are convenience-wrappers that make it easier to retrieve services from the registry.
    #

    @property
    def app(self) -> 'App':
        """Return App object."""
        return self.App

    @property
    def exit(self) -> 'Exit':
        """Return Exit object."""
        return self.Exit

    @property
    def inputs(self) -> 'Input':
        """Return an Inputs object."""
        return self.Input

    @property
    def logger(self) -> 'Logger':
        """Return an Inputs object."""
        return self.Logger

    @property
    def playbook(self) -> 'Playbook':
        """Return a Playbook object."""
        return self.app.playbook

    @property
    def redis_client(self) -> Redis:
        """Return a Redis client object (redis.Redis)."""
        return self.app.key_value_store.redis_client

    @property
    def request_tc(self) -> 'RequestsTc':
        """Return a TcSession."""
        return self.RequestsTc

    @property
    def session_tc(self) -> 'TcSession':
        """Return a TcSession."""
        return self.request_tc.session

    @property
    def token(self) -> 'Token':
        """Return a Token."""
        return self.app.token


registry = Registry()



"""TcEx Framework Module"""
# standard library
import gzip
import json
import tempfile
import uuid
from pathlib import Path
from typing import Literal

COMPRESS_LEVEl = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
MODE = Literal['w', 'wb', 'wt']


class FileOperation:
    """TcEx Utilities File Operations Class

    Args:
        out_path: The path of the defined "out" directory.
        temp_path: The path of the defined "temp" directory.
    """

    def __init__(
        self,
        out_path: Path | str | None = None,
        temp_path: Path | str | None = None,
    ):
        """Initialize the Class properties."""
        self.out_path = Path(out_path or tempfile.gettempdir())
        self.temp_path = Path(temp_path or tempfile.gettempdir())

    def _fqfn_out(self, filename: Path | str | None = None) -> Path:
        """Return a unique filename for the defined "out" directory."""
        # user provided filename or generate a unique one
        filename = filename if filename is not None else str(uuid.uuid4())

        # define the fully qualified path name
        return self.out_path / filename

    def _fqfn_temp(self, filename: Path | str | None = None) -> Path:
        """Return a unique filename for the defined "temp" directory."""
        # user provided filename or generate a unique one
        filename = filename if filename is not None else str(uuid.uuid4())

        # define the fully qualified path name
        return self.temp_path / filename

    @staticmethod
    def write_file(
        content: bytes | dict | list | str,
        fqfn: Path | str,
        mode: MODE = 'w',
        encoding: str | None = None,
        compress_level: COMPRESS_LEVEl | None = None,
    ) -> Path:
        """Write file content to a out directory, compressing if compress level provided.

        If passing binary data the mode needs to be set to 'wb'. If
        compress_level is provided mode is automatically set to 'wt'.

        Args:
            content: The file content.
            fqfn: A fully qualified file name.
            mode: The write mode ('w', 'wb', 'wt').
            encoding: The encoding to use when writing the file.
            compress_level: The compression level to use when writing the file.
        """
        # encoding is not supported when writing binary data
        encoding = encoding or 'utf-8'
        if mode == 'wb':
            encoding = None

        content = json.dumps(content) if isinstance(content, dict | list) else content
        fqfn = fqfn if isinstance(fqfn, Path) else Path(fqfn)

        # ensure output directory exists
        fqfn.parent.mkdir(parents=True, exist_ok=True)

        # write file, either normal or compressed
        if compress_level is not None:
            # overwrite mode if compressing
            mode = 'wb' if isinstance(content, bytes) else 'wt'
            with gzip.open(
                fqfn,
                mode,
                compresslevel=compress_level,
                encoding=encoding,
            ) as fh:
                fh.write(content)  # type: ignore
        else:
            with fqfn.open(mode, encoding=encoding) as fh:
                fh.write(content)

        # return the fully qualified path name
        return fqfn

    def write_out_binary_file(
        self,
        content: bytes,
        filename: str | None = None,
    ) -> Path:
        """Write content to a file in the defined "out" directory.

        Args:
            content: The file content.
            filename: The filename to use when writing the file.
        """
        return self.write_file(content, self._fqfn_out(filename), mode='wb')

    def write_out_compressed_file(
        self,
        content: bytes | dict | str,
        filename: Path | str | None = None,
        compress_level: COMPRESS_LEVEl = 9,
    ) -> Path:
        """Write content to a file in the defined "out" directory.

        Args:
            content: The file content.
            filename: The filename to use when writing the file.
            compress_level: The compression level to use when writing the file.
        """
        return self.write_file(
            content, self._fqfn_out(filename), mode='wt', compress_level=compress_level
        )

    def write_out_file(
        self,
        content: bytes | dict | str,
        filename: str | None = None,
        mode: MODE = 'w',
        encoding: str = 'utf-8',
        compress_level: COMPRESS_LEVEl | None = None,
    ) -> Path:
        """Write content to a file in the defined "out" directory.

        If passing binary data the mode needs to be set to 'wb'. If
        compress_level is provided mode is automatically set to 'wt'.

        Args:
            content: The file content.
            filename: A filename to use when writing the file.
            mode: The write mode ('w' or 'wb').
            encoding: The encoding to use when writing the file.
            compress_level: The compression level to use when writing the file.
        """
        # TODO: [high] fix this type ignore
        encoding = encoding if mode != 'wb' else None  # type: ignore
        return self.write_file(content, self._fqfn_out(filename), mode, encoding, compress_level)

    def write_temp_binary_file(
        self,
        content: bytes,
        filename: str | None = None,
    ) -> Path:
        """Write content to a file in the defined "temp" directory.

        Args:
            content: The file content.
            filename: The filename to use when writing the file.
        """
        return self.write_file(content, self._fqfn_temp(filename), mode='wb', encoding=None)

    def write_temp_compressed_file(
        self,
        content: bytes | dict | str,
        filename: str | None = None,
        compress_level: COMPRESS_LEVEl = 9,
    ) -> Path:
        """Write content to a file in the defined "temp" directory.

        Args:
            content: The file content.
            filename: The filename to use when writing the file.
            compress_level: The compression level to use when writing the file.
        """
        return self.write_file(
            content, self._fqfn_temp(filename), mode='wt', compress_level=compress_level
        )

    def write_temp_file(
        self,
        content: bytes | dict | str,
        filename: Path | str | None = None,
        mode: MODE = 'w',
        encoding: str = 'utf-8',
        compress_level: COMPRESS_LEVEl | None = None,
    ) -> Path:
        """Write content to a file in the defined "temp" directory.

        If passing binary data the mode needs to be set to 'wb'.

        Args:
            content: The file content.
            filename: The filename to use when writing the file.
            mode: The write mode ('w' or 'wb').
            encoding: The encoding to use when writing the file.
            compress_level: The compression level to use when writing the file.
        """
        # TODO: [high] fix this type ignore
        encoding = encoding if mode != 'wb' else None  # type: ignore
        return self.write_file(content, self._fqfn_temp(filename), mode, encoding, compress_level)



"""TcEx Framework Module"""

# third-party
from pydantic import Extra

# first-party
from tcex.input.model.api_model import ApiModel
from tcex.input.model.path_model import PathModel
from tcex.input.model.playbook_common_model import PlaybookCommonModel
from tcex.input.model.playbook_model import PlaybookModel
from tcex.input.model.proxy_model import ProxyModel
from tcex.input.model.service_model import ServiceModel


class ModuleAppModel(
    ApiModel, PathModel, PlaybookCommonModel, PlaybookModel, ProxyModel, ServiceModel
):
    """Model Definition

    This model provides all the inputs required by the "tcex.app" module.
    """

    class Config:
        """Model Config"""

        extra = Extra.ignore



"""TcEx Framework Module"""
# third-party
from requests import Session

# first-party
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.api.tc.v3.group_attributes.group_attribute import GroupAttribute, GroupAttributes
from tcex.api.tc.v3.groups.group import Group, Groups
from tcex.api.tc.v3.indicator_attributes.indicator_attribute import (
    IndicatorAttribute,
    IndicatorAttributes,
)
from tcex.api.tc.v3.indicators.indicator import Indicator, Indicators
from tcex.api.tc.v3.security_labels.security_label import SecurityLabel
from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset, VictimAssets
from tcex.api.tc.v3.victim_attributes.victim_attribute import VictimAttribute, VictimAttributes
from tcex.api.tc.v3.victims.victim import Victim, Victims


class ThreatIntelligence:
    """Threat Intelligence

    Args:
        session: An configured instance of request.Session with TC API Auth.
    """

    def __init__(self, session: Session):
        """Initialize instance properties."""
        self.session = session
        self._ti_utils = None

    @property
    def ti_utils(self):
        """Return instance of Threat Intel Utils."""
        if not self._ti_utils:
            self._ti_utils = ThreatIntelUtil(session_tc=self.session)
        return self._ti_utils

    def create_entity(self, entity: dict, owner: str) -> dict | None:
        """Create a CM object provided a dict and owner."""
        entity_type = entity['type'].lower()
        entity_type = entity_type.replace(' ', '_')
        try:
            if entity_type in (type_.lower() for type_ in self.ti_utils.group_types):
                main_type = 'Group'
                obj = self.group(**entity)
            elif entity_type in (type_.lower() for type_ in self.ti_utils.indicator_types):
                main_type = 'Indicator'
                obj = self.indicator(**entity)
            elif entity_type in ['victim']:
                main_type = 'Victim'
                obj = self.victim(**entity)
            else:
                raise RuntimeError(f'Invalid entity type provided for: {entity}')
        except AttributeError:
            return None

        r = obj.create()
        data: dict[str, int | str] = {'status_code': r.status_code}
        if r.ok:
            data.update(r.json().get('data', {}))
            data['main_type'] = main_type
            data['sub_type'] = entity_type
            data['owner'] = owner

        return data

    def group(self, **kwargs) -> Group:
        """Return a instance of Group object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            assignments (None, kwargs): A list of assignees and escalatees associated with this
                group (Task specific).
            associated_groups (Groups, kwargs): A list of groups associated with this group.
            associated_indicators (Indicators, kwargs): A list of indicators associated with this
                group.
            associated_victim_assets (VictimAssets, kwargs): A list of victim assets associated with
                this group.
            attributes (Attributes, kwargs): A list of Attributes corresponding to the Group.
            body (str, kwargs): The email Body.
            due_date (str, kwargs): The date and time that the Task is due.
            escalation_date (str, kwargs): The escalation date and time.
            event_date (str, kwargs): The date and time that the incident or event was first
                created.
            file_name (str, kwargs): The document or signature file name.
            file_text (str, kwargs): The signature file text.
            file_type (str, kwargs): The signature file type.
            first_seen (str, kwargs): The date and time that the campaign was first created.
            from_ (str, kwargs): The email From field.
            handles (AdversaryAssets, kwargs): A list of handle adversary assets associated with
                this group.
            header (str, kwargs): The email Header field.
            id (int, kwargs): The ID of the Group.
            malware (bool, kwargs): Is the document malware?
            name (str, kwargs): The name of the group.
            password (str, kwargs): The password associated with the document (Required if Malware
                is true).
            phone_numbers (AdversaryAssets, kwargs): A list of phone number adversary assets
                associated with this group.
            publish_date (str, kwargs): The date and time that the report was first created.
            reminder_date (str, kwargs): The reminder date and time.
            security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
                Intel item (NOTE: Setting this parameter will replace any existing tag(s) with the
                one(s) specified).
            status (str, kwargs): The status associated with this document, event, task, or incident
                (read only for task, document, and report).
            subject (str, kwargs): The email Subject section.
            tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this
                parameter will replace any existing tag(s) with the one(s) specified).
            to (str, kwargs): The email To field .
            type (str, kwargs): The **type** for the Group.
            urls (AdversaryAssets, kwargs): A list of url adversary assets associated with this
                group.
        """
        return Group(session=self.session, **kwargs)

    def group_attribute(self, **kwargs) -> GroupAttribute:
        """Return a instance of Group Attributes object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            default (bool, kwargs): A flag indicating that this is the default attribute of its type
                within the object. Only applies to certain attribute and data types.
            indicator_id (int, kwargs): Indicator associated with attribute.
            source (str, kwargs): The attribute source.
            type (str, kwargs): The attribute type.
            value (str, kwargs): Attribute value.
        """
        return GroupAttribute(session=self.session, **kwargs)

    def group_attributes(self, **kwargs) -> GroupAttributes:
        """Return a instance of Group Attributes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Group.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return GroupAttributes(session=self.session, **kwargs)

    def groups(self, **kwargs) -> Groups:
        """Return a instance of Groups object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Group.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return Groups(session=self.session, **kwargs)

    def indicator(self, **kwargs) -> Indicator:
        """Return a instance of Group object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            active (bool, kwargs): Is the indicator active?
            active_locked (bool, kwargs): Lock the indicator active value?
            address (str, kwargs): The email address associated with this indicator (EmailAddress
                specific summary field).
            associated_groups (Groups, kwargs): A list of groups that this indicator is associated
                with.
            associated_indicators (Indicators, kwargs): A list of indicators associated with this
                indicator.
            attributes (Attributes, kwargs): A list of Attributes corresponding to the Indicator.
            confidence (int, kwargs): The indicator threat confidence.
            description (str, kwargs): The indicator description text.
            dns_active (bool, kwargs): Is dns active for the indicator?
            host_name (str, kwargs): The host name of the indicator (Host specific summary field).
            ip (str, kwargs): The ip address associated with this indicator (Address specific
                summary field).
            md5 (str, kwargs): The md5 associated with this indicator (File specific summary field).
            private_flag (bool, kwargs): Is this indicator private?
            rating (int, kwargs): The indicator threat rating.
            security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
                Intel item (NOTE: Setting this parameter will replace any existing tag(s) with the
                one(s) specified).
            sha1 (str, kwargs): The sha1 associated with this indicator (File specific summary
                field).
            sha256 (str, kwargs): The sha256 associated with this indicator (File specific summary
                field).
            size (int, kwargs): The size of the file.
            source (str, kwargs): The source for this indicator.
            tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this
                parameter will replace any existing tag(s) with the one(s) specified).
            text (str, kwargs): The url text value of the indicator (Url specific summary field).
            type (str, kwargs): The **type** for the Indicator.
            value1 (str, kwargs): Custom Indicator summary field value1.
            value2 (str, kwargs): Custom Indicator summary field value2.
            value3 (str, kwargs): Custom Indicator summary field value3.
            whois_active (bool, kwargs): Is whois active for the indicator?
        """
        return Indicator(session=self.session, **kwargs)

    def indicator_attribute(self, **kwargs) -> IndicatorAttribute:
        """Return a instance of Case Attributes object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            default (bool, kwargs): A flag indicating that this is the default attribute of its type
                within the object. Only applies to certain attribute and data types.
            indicator_id (int, kwargs): Indicator associated with attribute.
            source (str, kwargs): The attribute source.
            type (str, kwargs): The attribute type.
            value (str, kwargs): Attribute value.
        """
        return IndicatorAttribute(session=self.session, **kwargs)

    def indicator_attributes(self, **kwargs) -> IndicatorAttributes:
        """Return a instance of Indicator Attributes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Group.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return IndicatorAttributes(session=self.session, **kwargs)

    def indicators(self, **kwargs) -> Indicators:
        """Return a instance of Indicators object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Indicator.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return Indicators(session=self.session, **kwargs)

    def security_label(self, **kwargs) -> SecurityLabel:
        """Return a instance of Case Attributes object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            color (str, kwargs): Color of the security label.
            description (str, kwargs): Description of the security label.
            name (str, kwargs): Name of the security label.
            owner (str, kwargs): The name of the Owner of the Label.
        """
        return SecurityLabel(session=self.session, **kwargs)

    def victim(self, **kwargs) -> Victim:
        """Return a instance of Victim object.

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            assets (VictimAssets, kwargs): A list of victim assets corresponding to the Victim.
            associated_groups (Groups, kwargs): A list of groups that this indicator is associated
                with.
            attributes (VictimAttributes, kwargs): A list of Attributes corresponding to the
                Victim.
            description (str, kwargs): The indicator description text.
            name (str, kwargs): Name of the Victim.
            nationality (str, kwargs): Nationality of the Victim.
            org (str, kwargs): Org of the Victim.
            id (int, kwargs): The ID of the Victim.
            owner_name (str, kwargs): The name of the Organization, Community, or Source that the
                item belongs to.
            security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
                Intel item (NOTE: Setting this parameter will replace any existing tag(s) with the
                one(s) specified).
            suborg (str, kwargs): Suborg of the Victim.
            tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this
                parameter will replace any existing tag(s) with the one(s) specified)
            type (str, kwargs): The type for the Victim.
            work_location (str, kwargs): Work Location of the Victim.
        """

        return Victim(session=self.session, **kwargs)

    def victims(self, **kwargs) -> Victims:
        """Return a instance of Victims object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Indicator.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return Victims(session=self.session, **kwargs)

    def victim_asset(self, **kwargs) -> VictimAsset:
        """Return a instance of VictimAsset object."""

        return VictimAsset(session=self.session, **kwargs)

    def victim_assets(self, **kwargs) -> VictimAssets:
        """Return a instance of Victims object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Indicator.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return VictimAssets(session=self.session, **kwargs)

    def victim_attribute(self, **kwargs) -> VictimAttribute:
        """Return a instance of VictimAttribute object."""

        return VictimAttribute(session=self.session, **kwargs)

    def victim_attributes(self, **kwargs) -> VictimAttributes:
        """Return a instance of Victims object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            **kwargs: Additional keyword arguments.

        Keyword Args:
            initial_response (dict, optional): Initial data in Case Object for Indicator.
            tql_filters (list, optional): A list of TQL filters.
            params (dict, optional): A dict of query params for the request.
        """
        return VictimAttributes(session=self.session, **kwargs)



"""TcEx Framework Module"""
# pylint: disable=no-self-argument
# standard library
import logging
import os
import platform
import re
import uuid
from enum import Enum
from importlib.metadata import version as get_version

# third-party
from pydantic import BaseModel, Field, validator
from pydantic.types import UUID4, UUID5, constr
from semantic_version import Version

__all__ = ['InstallJsonModel']

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


def snake_to_camel(snake_string: str) -> str:
    """Convert snake_case to camelCase"""
    components = snake_string.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


# define JSON encoders
json_encoders = {Version: str, UUID4: str, UUID5: str}


class _FeatureModel(BaseModel):
    """Model definition"""

    default: bool | None = Field(
        False, description='Indicates whether the feature is a default for the App type.'
    )
    runtime_levels: list[str] | None = Field(
        None,
        description='The runtime levels that the feature is valid for.',
    )
    version: Version | None = Field(
        None,
        description='The version of TcEx that the feature was added.',
    )

    class Config:
        """DataModel Config"""

        arbitrary_types_allowed = True
        json_encoders = json_encoders


class DeprecationModel(BaseModel):
    """Model definition for install_json.deprecation"""

    indicator_type: str | None = Field(
        None,
        description='The indicator type for the deprecation rule.',
    )
    interval_days: int | None = Field(
        None,
        description='The frequency the deprecation rule should run.',
    )
    confidence_amount: int | None = Field(
        None,
        description='The amount the confidence should be reduced by.',
    )
    delete_at_minimum: bool = Field(
        False,
        description='If true, the indicator will be deleted at the minimum confidence.',
    )
    percentage: bool = Field(
        False,
        description='If true, use percentage instead of point value when reducing the confidence.',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class FirstRunParamsModel(BaseModel):
    """Model definition for install_json.deprecation"""

    param: str | None = Field(
        None,
        description='The parameter to set to the first run value.',
    )
    value: int | str | None = Field(
        None,
        description='The value to set the parameter to.',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class FeedsModel(BaseModel):
    """Model definition for install_json.feeds"""

    attributes_file: str | None = Field(
        None,
        description=(
            'Optional property that provides the name of the CSV file with any custom '
            'Attributes required for the feed (e.g., attribute.json).'
        ),
    )
    deprecation: list[DeprecationModel] = Field(
        [],
        description='The deprecation rules for the feed.',
    )
    document_storage_limit_mb: int = Field(
        ...,
        description='Optional property that sets the Document storage limit.',
    )
    enable_bulk_json: bool = Field(
        False,
        description='Optional property that enables or disables the bulk JSON capability.',
    )
    first_run_params: list[FirstRunParamsModel] = Field(
        [],
        description='Param overrides for the first run of the feed.',
    )
    indicator_limit: int = Field(
        ...,
        description='Optional property that sets the Indicator limit.',
    )
    job_file: str = Field(
        ...,
        description=(
            'Optional property that provides the name of the JSON file that is used to '
            'set up and run the Job that pulls in content from the feed.'
        ),
    )
    source_category: str = Field(
        ...,
        description='Optional property that specifies how the source should be categorized.',
    )
    source_description: str = Field(
        ...,
        description=(
            '''Optional property that provides the source's description as it will be '''
            '''displayed in the ThreatConnect platform.'''
        ),
    )
    source_name: str = Field(
        ...,
        description=(
            '''Optional property that provides the name of the source in which the feed's '''
            '''content will be created.'''
        ),
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class ExposePlaybookKeyAsEnum(str, Enum):
    """Enum for install_json.params[].exposePlaybookAs"""

    Binary = 'Binary'
    BinaryArray = 'BinaryArray'
    KeyValue = 'KeyValue'
    KeyValueArray = 'KeyValueArray'
    String = 'String'
    StringArray = 'StringArray'
    TCEntity = 'TCEntity'
    TCEntityArray = 'TCEntityArray'


class TypeEnum(str, Enum):
    """Enum for install_json.params[].type"""

    Boolean = 'Boolean'
    Choice = 'Choice'
    EditChoice = 'EditChoice'
    KeyValueList = 'KeyValueList'
    MultiChoice = 'MultiChoice'
    String = 'String'
    StringMixed = 'StringMixed'


class ParamsModel(BaseModel):
    """Model definition for install_json.params"""

    allow_multiple: bool = Field(
        False,
        description=(
            'The value of this optional property is automatically set to true if the '
            'MultiChoice type is used. If a String type is used, this flag allows the '
            'user to define multiple values in a single input field delimited by a pipe '
            '("|") character.'
        ),
    )
    allow_nested: bool = Field(
        False,
        description='',
    )
    default: bool | str | None = Field(
        None,
        description='Optional property that is the default value for an App input parameter.',
    )
    encrypt: bool = Field(
        False,
        description=(
            'Optional property that designates a parameter as an encrypted value. '
            'Parameters defined as encrypted will be managed by the Keychain feature '
            'that encrypts password while at rest. This flag should be used with the '
            'String type and will render a password input text box in the App '
            'configuration.'
        ),
    )
    expose_playbook_key_as: ExposePlaybookKeyAsEnum | None = Field(
        None,
        description='',
    )
    feed_deployer: bool = Field(
        False,
        description='',
    )
    hidden: bool = Field(
        False,
        description=(
            'If this optional property is set to true, this parameter will be hidden '
            'from the Job Wizard. Hidden parameters allow the developer to persist '
            'parameters between Job executions without the need to render the values in '
            'the Job Wizard. This option is valid only for Python and Java Apps. Further '
            'details on persisting parameters directly from the app are beyond the scope '
            'of this documentation.'
        ),
    )
    intel_type: list[str] = Field(
        [],
        description='',
    )
    label: str = Field(
        ...,
        description=(
            'Required property providing a description of the parameter displayed in the '
            'Job Wizard or Spaces Configuration dialog box within the ThreatConnect '
            'platform.'
        ),
    )
    name: str = Field(
        ...,
        description=(
            'Required property that is the internal parameter name taken from the Job '
            'Wizard and passed to the App at runtime. It is the effective command-line '
            'argument name passed to the App.'
        ),
    )
    note: str | None = Field(
        None,
        description=(
            'Optional parameter-description field available in Playbook Apps under the ? '
            'tooltip when the App parameters are being edited. Use this field to '
            'describe the purpose of the parameter in two to three sentences.'
        ),
    )
    playbook_data_type: list[str] = Field(
        [],
        description=(
            'Optional property restricting the data type of incoming Playbook variables. '
            'This is different than the type property that controls the UI input type. '
            'The playbookDataType can be any standard or custom type and is expected to '
            'be an array of strings.'
        ),
    )
    required: bool = Field(
        False,
        description=(
            'Optional property designating this parameter as a required field that must '
            'be populated to save the Job or Playbook App.'
        ),
    )
    sequence: int | None = Field(
        None,
        description=(
            'Optional number used to control the ordering of the parameters in the Job '
            'Wizard or Spaces Configuration dialog box. If it is not defined, the order '
            'of the parameters in the install.json file will be used.'
        ),
    )
    service_config: bool = Field(
        False,
        description='',
    )
    setup: bool = Field(
        False,
        description='',
    )
    type: TypeEnum = Field(
        ...,
        description=(
            'Required property to enable the UI to display relevant components and allow '
            'the Job Executor to adapt how parameters are passed to an App at runtime. '
            'The table below lists the available types and how they affect elements '
            'within the platform.'
        ),
    )
    valid_values: list[str] = Field(
        [],
        description=(
            'Optional property to be used with the Choice, MultiChoice, and String input '
            'types to provide pre-defined inputs for the user selection.'
        ),
    )
    view_rows: int | None = Field(
        None,
        description=(
            'Optional property for Playbook Apps to control the height of the display in '
            'the input parameter, and it expects an integer value. A value of 1 is '
            'default (and will show a text input element) and anything greater than 1 '
            'displays a textarea input when editing the Playbook App in ThreatConnect.'
        ),
    )

    @validator('name')
    def _name(cls, v):
        """Return the transformed "name" field.

        Used to replace labels for fields non-alphanumeric chars (migrate label to name).
        """
        if v is not None:
            v = v.lower().replace(' ', '_')

            # remove all non-alphanumeric characters and underscores
            v = re.sub(r'[^a-zA-Z0-9_]', '', v)
        return v

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        smart_union = True
        use_enum_values = True
        validate_assignment = False

    def __hash__(self):
        """Make model hashable."""
        return hash(self.name)


class OutputVariablesModel(BaseModel):
    """Model definition for install_json.playbook.outputVariables"""

    # sensitive value
    encrypt: bool = Field(
        False,
        description='',
    )
    intel_type: list | None = Field(
        None,
        description='',
    )
    name: str = Field(
        ...,
        description='',
    )
    note: str | None = Field(
        None,
        description='',
    )
    type: str = Field(
        ..., description='Required property that specifies the type of the output variable.'
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True

    def __hash__(self):
        """Make model hashable."""
        return hash(f'{self.name}{self.type}')


class RetryModel(BaseModel):
    """Model definition for install_json.playbook.retry"""

    actions: list[str] = Field(
        [],
        description='A list of tc_actions that support retry.',
    )
    allowed: bool = Field(
        False,
        description=(
            'Optional property that specifies whether the Playbook App can retry its ' 'execution.'
        ),
    )
    default_delay_minutes: int = Field(
        ...,
        description=(
            'Optional property that specifies the number of minutes between each new '
            'retry in case of failure. This property assumes that the allowed property '
            'is set to true to allow the App to retry.'
        ),
    )
    default_max_retries: int = Field(
        ...,
        description=(
            'Optional property that specifies the maximum number of times the Playbook '
            'App can retry in case of failure. This property assumes that the allowed '
            'property is set to true to allow the app to retry.'
        ),
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class PlaybookModel(BaseModel):
    """Model definition for install_json.playbook"""

    output_prefix: str | None = Field(None, description='')
    output_variables: list[OutputVariablesModel] = Field(
        [],
        description=(
            'Optional outputVariables property that specifies the variables that a '
            'Playbook App will provide for downstream Playbooks.'
        ),
    )
    retry: RetryModel | None = Field(
        None,
        description=(
            'Optional retry property that can be used to allow a Playbook to retry its '
            'execution in case of failure.'
        ),
    )
    type: str = Field(
        ...,
        description='The App category (e.g., Endpoint Detection and Response).',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class ServiceModel(BaseModel):
    """Model definition for install_json.service"""

    discovery_types: list[str] = Field(
        [],
        description='Service App discovery types (e.g., TaxiiApi).',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


def get_commit_hash() -> str | None:
    """Return the current commit hash if available.

    This is not a required task so best effort is fine. In other words this is not guaranteed
    to work 100% of the time.
    """
    commit_hash = None
    branch = None
    branch_file = '.git/HEAD'  # ref: refs/heads/develop

    # get current branch
    if os.path.isfile(branch_file):
        with open(branch_file, encoding='utf-8') as f:
            try:
                branch = '/'.join(f.read().strip().split('/')[2:])
            except IndexError:  # pragma: no cover
                pass

        # get commit hash
        if branch:
            hash_file = f'.git/refs/heads/{branch}'
            if os.path.isfile(hash_file):
                with open(hash_file, encoding='utf-8') as f:
                    commit_hash = f.read().strip()

    if commit_hash is None:
        # gitlab / github CI environment variable
        commit_hash = os.getenv('CI_COMMIT_SHA') or os.getenv('GITHUB_SHA')

    return commit_hash


def gen_app_id() -> UUID5:
    """Return a generate id for the current App."""
    return uuid.uuid5(uuid.NAMESPACE_X500, os.path.basename(os.getcwd()).lower())


class InstallJsonCommonModel(BaseModel):
    """Model definition for install.json common configuration

    This model contains the common fields for the install.json file and
    the app_spec.yaml file.
    """

    allow_on_demand: bool = Field(
        False,
        description=(
            'Required property that allows or disallows an App to be run on demand using '
            'the Run Now button when the App is configured as a Job in the ThreatConnect '
            'platform. This property only applies to Python and Java Apps.'
        ),
    )
    allow_run_as_user: bool = Field(
        False,
        description='Controls whether a Playbook App supports run-as-users.',
    )
    api_user_token_param: bool = Field(
        True,
        description=(
            '[Deprecated] Optional property that specifies whether or not the App should '
            'use an API user token (which allows access to the DataStore).'
        ),
    )
    app_id: UUID4 | UUID5 = Field(
        default_factory=gen_app_id,
        description=(
            '[TcEx 1.1.4+] A unique identifier for the App. This field is not currently '
            'used in the core product, but will be used in other tooling to identify the '
            'App. The appId field with the major version from programVersion make up a '
            'unique Application release. If this field does not exist while packaging '
            'the App via the `tcex package` command, a value will be added using the '
            'project directory name as a seed. Once an App has been released the appId '
            'field should not be changed.'
        ),
    )
    category: str = Field(
        '',
        description='The category of the App. Also playbook.type for playbook Apps.',
    )
    deprecates_apps: list[str] = Field(
        [],
        description=(
            'Optional property that provides a list of Apps that should be '
            'deprecated by this App.'
        ),
    )
    display_name: constr(min_length=3, max_length=100) = Field(  # type: ignore
        ...,
        description=(
            'Required property providing the name of the App as it will be displayed in '
            'the ThreatConnect platform.'
        ),
    )
    display_path: constr(min_length=3, max_length=100) | None = Field(  # type: ignore
        None,
        description='The display path for API service Apps.',
    )
    features: list[str] = Field(
        ...,
        description=(
            'An array of supported features for the App. These feature enable '
            'additional functionality in the Core Platform and/or for the App.'
        ),
    )
    labels: list[str] = Field(
        [],
        description='A list of labels for the App.',
    )
    language_version: str = Field(
        ...,
        description=(
            'The major.minor version of the language (e.g., Python "3.11"). This value is used by '
            'the Core platform to control which version of Python is used to launch the App.'
        ),
    )
    list_delimiter: str = Field(
        ...,
        description=(
            'Optional property that sets the character used to delimit the values of '
            'an input that support the allowMultiple param option.'
        ),
    )
    min_server_version: Version = Field(
        '7.2.0',
        description=(
            'Optional string property restricting the ThreatConnect instance from '
            'installing the App if it does not meet this version requirement (e.g., 7.2.0).'
        ),
    )
    note: str | None = Field(
        None,
        description=(
            'Optional property available in Playbook Apps while configuring App inputs '
            'in the UI. This is the top level not of the App and should describe the '
            'functionality and use cases of the App.'
        ),
    )
    program_language: str = Field(
        ...,
        description=(
            'Required property describing the language runtime environment used by the '
            'ThreatConnect Job Executor. It is relevant for Apps that run on the Job '
            'Execution Engine (Python and Java Apps) and can be set to NONE for Spaces '
            'Apps.'
        ),
    )
    program_main: str = Field(
        ...,
        description=(
            'Required property providing the entry point into the App. For Python Apps, '
            'it is the name of the .py file (or exclude the extension if running it as a '
            'module). For Java Apps, it is the main class the Job Execution Engine '
            'should use when calling the App using the Java Runtime Environment.'
        ),
    )
    program_version: Version = Field(
        ...,
        description=(
            'Required property providing the version number for the App that will be '
            'displayed in the Installed Apps section available to a System '
            'Administrator. ThreatConnect recommends the use of semantic versioning '
            '(e.g., 1.0.1).'
        ),
    )
    runtime_level: str = Field(
        ...,
        description='The type for the App (e.g., Playbook, Organization, etc).',
    )
    service: ServiceModel | None = Field(
        None,
        description='',
    )
    sdk_version: Version | None = Field(
        None,
        description=(
            'The version of the SDK (TcEx). This value is used by the Core '
            'platform to control behavior in App Builder.'
        ),
    )

    @validator('language_version', always=True, pre=True)
    def _language_version(cls, v) -> str:
        """Return a version object for "version" field."""

        def _major_minor(v):
            """Return the major.minor version."""
            try:
                version_ = Version.coerce(v)
                v = f'{version_.major}.{version_.minor}'
            except Exception:  # nosec
                # best effort
                pass

            return v

        if v is None:
            v = _major_minor(platform.python_version())
        else:
            # for TC version >= 7.0.1 and < 7.2.0
            v = _major_minor(v)

        return v

    @validator('min_server_version', pre=True)
    def _min_server_version(cls, v) -> Version:
        """Return a version object for "version" fields."""
        # all tcex 4 Apps must have min server version of at least 7.2.0
        if v is None or Version.coerce(v) < Version('7.2.0'):
            v = '7.2.0'
        return Version.coerce(v)

    @validator('program_version', pre=True)
    def _program_version(cls, v) -> Version:
        """Return a version object for "version" fields."""
        if v is not None:
            return Version(v)
        return v

    @validator('sdk_version', always=True, pre=True)
    def _sdk_version(cls, v) -> Version:
        """Return a version object for "version" field."""
        if v is None:
            # assume legacy App
            return Version('2.0.0')

        # ensure v is a Version object
        v = v if isinstance(v, Version) else Version.coerce(v)

        # update version
        if v >= Version('4.0.0'):
            try:
                # best effort to update the tcex version
                return Version.coerce(get_version('tcex'))
            except Exception:
                return v

        return v

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        arbitrary_types_allowed = True
        json_encoders = json_encoders
        validate_assignment = True

    @property
    def is_api_service_app(self) -> bool:
        """Return True if the current App is ANY type of API Service App."""
        return self.runtime_level.lower() in [
            'apiservice',
            'feedapiservice',
        ]

    @property
    def is_feed_app(self) -> bool:
        """Return True if the current App is ANY type of API Service App."""
        return self.runtime_level.lower() in [
            'feedapiservice',
            'organization',
        ]

    @property
    def is_job_app(self) -> bool:
        """Return True if the current App is an Organization (job) App."""
        return self.is_organization_app

    @property
    def is_organization_app(self) -> bool:
        """Return True if the current App is an Organization (job) App."""
        return self.runtime_level.lower() == 'organization'

    @property
    def is_playbook_app(self) -> bool:
        """Return True if the current App is a Playbook App."""
        return self.runtime_level.lower() == 'playbook'

    @property
    def is_trigger_app(self) -> bool:
        """Return True if the current App is trigger Service App."""
        return self.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']

    @property
    def is_webhook_trigger_app(self) -> bool:
        """Return True if the current App is a webhook Service App."""
        return self.runtime_level.lower() == 'webhooktriggerservice'

    @property
    def is_service_app(self) -> bool:
        """Return True if the current App is ANY type of Service App."""
        return self.runtime_level.lower() in [
            'apiservice',
            'feedapiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]

    @property
    def known_features(self) -> dict[str, _FeatureModel]:
        """Return all known features."""

        feature_data = {
            'advancedRequest': {
                'runtime_levels': ['playbook'],
            },
            'aotExecutionEnabled': {
                'default': True,
                'runtime_levels': ['playbook'],
            },
            'appBuilderCompliant': {
                'default': True,
                'runtime_levels': ['playbook', 'triggerservice', 'webhooktriggerservice'],
            },
            'CALSettings': {},
            'fileParams': {
                'default': True,
            },
            # 'layoutEnabledApp': {
            #     'runtime_levels': ['playbook', 'triggerservice', 'webhooktriggerservice'],
            # },
            'linkApiPath': {
                'runtime_levels': ['apiservice', 'feedapiservice'],
            },
            'redisPasswordSupport': {
                'default': True,
                'version': Version('3.0.9'),
            },
            'runtimeVariables': {
                'default': True,
                'version': Version('3.0.2'),
            },
            # 'secureParams': {
            #     'default': True,
            #     'runtime_levels': ['organization', 'playbook'],
            # },
            'smtpSettings': {},
            'webhookResponseMarshall': {
                'runtime_levels': ['webhooktriggerservice'],
                'version': Version('4.0.0'),
            },
            'webhookServiceEndpoint': {
                'runtime_levels': ['webhooktriggerservice'],
                'version': Version('4.0.0'),
            },
            # features for TC App loop prevention
            'CreatesGroup': {
                'runtime_levels': ['playbook'],
            },
            'CreatesIndicator': {
                'runtime_levels': ['playbook'],
            },
            'CreatesSecurityLabel': {
                'runtime_levels': ['playbook'],
            },
            'CreatesTag': {
                'runtime_levels': ['playbook'],
            },
            'DeletesGroup': {
                'runtime_levels': ['playbook'],
            },
            'DeletesIndicator': {
                'runtime_levels': ['playbook'],
            },
            'DeletesSecurityLabel': {
                'runtime_levels': ['playbook'],
            },
            'DeletesTag': {
                'runtime_levels': ['playbook'],
            },
        }
        return {k: _FeatureModel(**v) for k, v in feature_data.items()}

    @property
    def updated_features(self) -> list[str]:
        """Update feature set based on App type."""
        try:
            tcex_version = Version.coerce(get_version('tcex'))
        except Exception:
            tcex_version = Version('2.0.0')

        # normalize features based on App type and TcEx version
        features = []
        for feature, model in self.known_features.items():
            if (
                model.default is True
                and (model.version is None or model.version <= tcex_version)
                and (
                    model.runtime_levels is None
                    or self.runtime_level.lower() in model.runtime_levels
                )
            ):
                features.append(feature)

        # add layoutEnabledApp if layout.json file exists in project
        if os.path.isfile('layout.json'):
            features.append('layoutEnabledApp')

        # extend feature list with features defined by developer
        for feature in self.features:
            model = self.known_features.get(feature)

            if model is None:
                # not sure what the feature is, but add it anyway
                features.append(feature)

                # log unknown features
                _logger.warning(f'Unknown feature found in install.json: {feature}')
            else:
                # "drop" features that should not be in the list
                if (model.version is None or model.version <= tcex_version) and (
                    model.runtime_levels is None
                    or self.runtime_level.lower() in model.runtime_levels
                ):
                    features.append(feature)

        return sorted(set(features))


class InstallJsonOrganizationModel(BaseModel):
    """Install JSON Common Model

    This model contains the common fields for the install.json file and
    the app_spec.yaml file.
    """

    feeds: list[FeedsModel] = Field(
        [],
        description='A list of features enabled for the App.',
    )
    publish_out_files: list[str] = Field(
        [],
        description=(
            'Optional field available for job-style Apps that can be scheduled to serve '
            'files. If this array is populated, the App is responsible for writing the '
            'files to the relative tc_output_path parameter that is passed in. This will '
            'enable HTTP-based file serving of these files as a unique URL available to '
            'the user in ThreatConnect. This parameter accepts an array of strings and '
            'can include file globs.'
        ),
    )
    repeating_minutes: list[int] = Field(
        [],
        description=(
            'Optional property that provides a list of minute increments to display in '
            'the Repeat Every… section in the Schedule panel of the Job Wizard. This '
            'property is relevant only for Python and Java Apps for which the developer '
            'wants to control how frequently an App can be executed. If this property is '
            'not defined, the default listing is as follows: [60, 120, 240, 360, 720].'
        ),
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class InstallJsonModel(InstallJsonCommonModel, InstallJsonOrganizationModel):
    """Model definition for install.json configuration file"""

    commit_hash: str | None = Field(
        default_factory=get_commit_hash,
        description='The git commit hash from when the App was built.',
    )
    docker_image: str | None = Field(
        None,
        description='[unsupported] The docker image to run the App.',
    )
    params: list[ParamsModel] = Field(
        [],
        description='',
    )
    playbook: PlaybookModel | None = Field(
        None,
        description='',
    )
    program_icon: str | None = Field(
        None,
        description=(
            'Optional property providing the icon that will be used to represent Central '
            'Spaces Apps.'
        ),
    )
    program_name: str | None = Field(
        None,
        description='',
    )
    runtime_context: list = Field(
        [],
        description=(
            'Optional property enabling Spaces Apps to be context aware (i.e., Spaces '
            'Apps that can be added to the Details screen of an object in the '
            'ThreatConnect platform). Because this property is an array of strings, the '
            'App can be displayed in Spaces under multiple contexts within the '
            'ThreatConnect platform, including the Menu and Search screens. This property '
            'is only applicable to Spaces Apps.'
        ),
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        arbitrary_types_allowed = True
        json_encoders = json_encoders
        validate_assignment = True

    @property
    def app_output_var_type(self) -> str:
        """Return the appropriate output var type for the current App."""
        if self.is_trigger_app:
            return 'Trigger'
        return 'App'

    def filter_params(
        self,
        name: str | None = None,
        hidden: bool | None = None,
        required: bool | None = None,
        service_config: bool | None = None,
        _type: str | None = None,
        input_permutations: dict | None = None,
    ) -> dict[str, ParamsModel]:
        """Return params as name/data dict.

        Args:
            name: The name of the input to return.
            hidden: If set the inputs will be filtered based on hidden field.
            required: If set the inputs will be filtered based on required field.
            service_config: If set the inputs will be filtered based on serviceConfig field.
            _type: The type of input to return.
            input_permutations: A list of valid input names for provided permutation.

        Returns:
            dict: All valid inputs for current filter.
        """
        params = {}
        for p in self.params:
            if name is not None:
                if p.name != name:
                    continue

            if hidden is not None:
                if p.hidden is not hidden:
                    continue

            if required is not None:
                if p.required is not required:
                    continue

            if service_config is not None:
                if p.service_config is not service_config:
                    continue

            if _type is not None:
                if p.type != _type:
                    continue

            if input_permutations is not None:
                if p.name not in input_permutations:
                    continue

            params.setdefault(p.name, p)
        return params

    def get_output(self, name: str) -> OutputVariablesModel | None:
        """Return output for the matching name."""
        return self.playbook_outputs.get(name)

    def get_param(self, name: str) -> ParamsModel | None:
        """Return param for the matching name."""
        return self.params_dict.get(name)

    @property
    def optional_params(self) -> dict[str, ParamsModel]:
        """Return params as name/data model."""
        return {p.name: p for p in self.params if p.required is False}

    @property
    def package_version(self):
        """Return the major version of the App."""
        return f'v{self.program_version.major}'

    @property
    def param_names(self) -> list[str]:
        """Return the "name" field from all params."""
        return [p.name for p in self.params]

    @property
    def params_dict(self) -> dict[str, ParamsModel]:
        """Return params as name/data dict."""
        return {p.name: p for p in self.params}

    @property
    def playbook_outputs(self) -> dict[str, OutputVariablesModel]:
        """Return outputs as name/data model."""
        return {} if self.playbook is None else {o.name: o for o in self.playbook.output_variables}

    @property
    def required_params(self) -> dict[str, ParamsModel]:
        """Return params as name/data dict."""
        return {p.name: p for p in self.params if p.required is True}

    @property
    def service_config_params(self) -> dict[str, ParamsModel]:
        """Return params as name/data dict."""
        return {p.name: p for p in self.params if p.service_config is True}

    @property
    def service_playbook_params(self) -> dict[str, ParamsModel]:
        """Return params as name/data dict."""
        return {p.name: p for p in self.params if p.service_config is False}



"""TcEx Framework Module"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class UserGroupFilter(FilterABC):
    """Filter Object for UserGroups"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.USER_GROUPS.value

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the user group.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the user group.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the user group.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)



"""TcEx Framework Module"""

# first-party
from tcex.api.tc.v3.attribute_types.attribute_type import AttributeType, AttributeTypes
from tcex.api.tc.v3.case_management.case_management import CaseManagement
from tcex.api.tc.v3.security.security import Security
from tcex.api.tc.v3.threat_intelligence.threat_intelligence import ThreatIntelligence


class V3(CaseManagement, Security, ThreatIntelligence):
    """V3 API Collection

    Args:
        session: An configured instance of request.Session with TC API Auth.
    """

    def attribute_type(self, **kwargs) -> AttributeType:
        """Return a instance of Attribute Types object."""
        return AttributeType(session=self.session, **kwargs)

    def attribute_types(self, **kwargs) -> AttributeTypes:
        """Return a instance of Attribute Types object."""
        return AttributeTypes(session=self.session, **kwargs)

    @property
    def cm(self) -> CaseManagement:
        """Return Case Management API collection."""
        return CaseManagement(self.session)

    @property
    def security(self) -> Security:
        """Return Security API collection."""
        return Security(self.session)

    @property
    def ti(self) -> ThreatIntelligence:
        """Return Threat Intelligence API collection."""
        return ThreatIntelligence(self.session)



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class AttributeTypeModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='AttributeType Model',
    validate_assignment=True,
):
    """Attribute_Type Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    allow_markdown: bool = Field(
        None,
        description='Flag that enables markdown feature in the attribute value field.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='allowMarkdown',
    )
    description: str | None = Field(
        None,
        description='The description of the attribute type.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    error_message: str | None = Field(
        None,
        description='The error message displayed.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='errorMessage',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    max_size: int | None = Field(
        None,
        description='The maximum size of the attribute value.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='maxSize',
    )
    name: str | None = Field(
        None,
        description='The name of the attribute type.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    validation_rule: dict | None = Field(
        None,
        description='The validation rule that governs the attribute value.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='validationRule',
    )


class AttributeTypeDataModel(
    BaseModel,
    title='AttributeType Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Attribute_Types Data Model"""

    data: list[AttributeTypeModel] | None = Field(
        [],
        description='The data for the AttributeTypes.',
        methods=['POST', 'PUT'],
        title='data',
    )


class AttributeTypesModel(
    BaseModel,
    title='AttributeTypes Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Attribute_Types Model"""

    _mode_support = PrivateAttr(False)

    data: list[AttributeTypeModel] | None = Field(
        [],
        description='The data for the AttributeTypes.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
AttributeTypeDataModel.update_forward_refs()
AttributeTypeModel.update_forward_refs()
AttributeTypesModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
import logging
from abc import ABC
from collections.abc import Generator
from typing import Any

# third-party
from requests import Response, Session
from requests.exceptions import ProxyError, RetryError

# first-party
from tcex.api.tc.v3.tql.tql import Tql
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.cached_property import cached_property
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class ObjectCollectionABC(ABC):
    """Case Management Collection Abstract Base Class

    This class is a base class for Case Management collections that use
    multi-inheritance with a pydantic BaseModel class. To ensure
    properties are not added to the model both @property and @setter
    methods are used.
    """

    def __init__(
        self,
        session: Session,
        tql_filters: list | None = None,  # This will be removed!
        params: dict | None = None,
    ):
        """Initialize instance properties."""
        self._params = params or {}
        self._tql_filters = tql_filters or []

        # properties
        self._session = session
        self.log = _logger
        self.request: Response
        self.tql = Tql()
        self._model = None
        self.type_ = None  # defined in child class
        self.util = Util()

    def __len__(self) -> int:
        """Return the length of the collection."""
        parameters = self._params.copy()
        parameters['resultLimit'] = 1
        parameters['count'] = True
        tql_string = self.tql.raw_tql
        if not self.tql.raw_tql:
            tql_string = self.tql.as_str
        if tql_string:
            parameters['tql'] = tql_string

        # convert all keys to camel case
        for k, v in list(parameters.items()):
            k = self.util.snake_to_camel(k)
            # if result_limit and resultLimit both show up use the proper cased version
            if k not in parameters:
                parameters[k] = v

        self._request(
            'GET',
            self._api_endpoint,
            body=None,
            params=parameters,
            headers={'content-type': 'application/json'},
        )
        return self.request.json().get('count', len(self.request.json().get('data', [])))

    @property
    def _api_endpoint(self):  # pragma: no cover
        """Return filter method."""
        raise NotImplementedError('Child class must implement this method.')

    def _request(
        self,
        method: str,
        url: str,
        body: bytes | str | None = None,
        params: dict | None = None,
        headers: dict | None = None,
    ):
        """Handle standard request with error checking."""
        try:
            self.request = self._session.request(
                method, url, data=body, headers=headers, params=params
            )
            self.log.debug(f'feature=api-tc-v3, request-body={self.request.request.body}')
        except (ConnectionError, ProxyError, RetryError):  # pragma: no cover
            handle_error(
                code=951,
                message_values=[
                    method.upper(),
                    None,
                    '{\"message\": \"Connection/Proxy Error/Retry\"}',
                    url,
                ],
            )

        if not self.success(self.request):
            err = self.request.text or self.request.reason
            handle_error(
                code=950,
                message_values=[
                    self.request.request.method,
                    self.request.status_code,
                    err,
                    self.request.url,
                ],
            )

        # log content for debugging
        self.log_response_text(self.request)

    @property
    def filter(self):  # pragma: no cover
        """Return filter method."""
        raise NotImplementedError('Child class must implement this method.')

    def log_response_text(self, response: Response):
        """Log the response text."""
        response_text = 'response text: (text to large to log)'
        if len(response.content) < 5000:  # check size of content for performance
            response_text = response.text
        self.log.debug(f'feature=api-tc-v3, response-body={response_text}')

    @property
    def model(self):
        """Return the model."""
        return self._model

    @model.setter
    def model(self, data):
        self._model = type(self.model)(**data)

    def iterate(
        self,
        base_class: Any,
        api_endpoint: str | None = None,
        params: dict | None = None,
    ) -> Generator:
        """Iterate over CM/TI objects."""
        url = api_endpoint or self._api_endpoint
        params = params or self.params

        # special parameter for indicators to enable the return the the indicator fields
        # (value1, value2, value3) on std-custom/custom-custom indicator types.
        if self.type_ == 'Indicators' and api_endpoint is None:
            params.setdefault('fields', []).append('genericCustomIndicatorValues')

        # convert all keys to camel case
        for k, v in list(params.items()):
            k = self.util.snake_to_camel(k)
            params[k] = v

        tql_string = self.tql.raw_tql or self.tql.as_str

        if tql_string:
            params['tql'] = tql_string

        while True:
            self._request(
                'GET',
                body=None,
                url=url,
                headers={'content-type': 'application/json'},
                params=params,
            )

            # reset some vars
            params = {}

            response = self.request.json()
            data = response.get('data', [])
            url = response.pop('next', None)

            for result in data:
                yield base_class(session=self._session, **result)  # type: ignore

            # break out of pagination if no next url present in results
            if not url:
                break

    @property
    def params(self) -> dict:
        """Return the parameters of the case management object collection."""
        return self._params

    @params.setter
    def params(self, params: dict):
        """Set the parameters of the case management object collection."""
        self._params = params

    @staticmethod
    def success(r: Response) -> bool:
        """Validate the response is valid.

        Args:
            r (requests.response): The response object.

        Returns:
            bool: True if status is "ok"
        """
        status = True
        if r.ok:
            try:
                if r.json().get('status') != 'Success':  # pragma: no cover
                    status = False
            except Exception:  # pragma: no cover
                status = False
        else:
            status = False
        return status

    @property
    def timeout(self) -> int:
        """Return the timeout of the case management object collection."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: int):
        """Set the timeout of the case management object collection."""
        self._timeout = timeout

    @cached_property
    def tql_options(self):
        """Return TQL data keywords."""
        _data = []
        r = self._session.options(f'{self._api_endpoint}/tql', params={})
        if r.ok:
            _data = r.json()['data']
        return _data

    @property
    def tql_keywords(self):
        """Return supported TQL keywords."""
        return [to.get('keyword') for to in self.tql_options]



"""TcEx Framework Module"""
# standard library
import logging
from functools import cached_property

# first-party
from tcex.app.config.install_json import InstallJson
from tcex.input.model.module_requests_session_model import ModuleRequestsSessionModel
from tcex.registry import registry
from tcex.requests_external.external_session import ExternalSession

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class RequestsExternal:
    """Requests External"""

    def __init__(
        self,
        model: ModuleRequestsSessionModel,
        proxies: dict[str, str],
        user_agent: dict[str, str] | None = None,
    ):
        """Initialize instance properties."""
        self.model = model
        self.proxies = proxies
        self.user_agent = user_agent or registry.app.user_agent

        # properties
        self.install_json = InstallJson()
        self.log = _logger

    def get_session(self, log_curl: bool = True) -> ExternalSession:
        """Return an instance of Requests Session configured for External API requests."""
        _session_external = ExternalSession()

        # add User-Agent to headers
        if self.user_agent:
            _session_external.headers.update(self.user_agent)

        # add proxy support if requested
        if self.model.tc_proxy_external:
            _session_external.proxies = self.proxies
            self.log.info(
                f'Using proxy host {self.model.tc_proxy_host}:'
                f'{self.model.tc_proxy_port} for external session.'
            )

        if self.model.tc_log_curl:
            _session_external.log_curl = log_curl

        return _session_external

    @cached_property
    def session(self) -> ExternalSession:
        """Return an instance of Requests Session configured for the ThreatConnect API."""
        return self.get_session()



"""TcEx Framework Module"""
# standard library
import random
import re
from functools import cached_property, reduce
from string import ascii_letters

# third-party
import inflect


class StringOperation:
    """TcEx Utilities String Operations Class"""

    @staticmethod
    def camel_string(string: str) -> 'CamelString':
        """Return str with custom properties/methods."""
        return CamelString(string)

    def camel_to_snake(self, string: str) -> str:
        """Return snake_case string from a camelCase string."""
        return self._camel_pattern.sub('_', string).lower()

    def camel_to_space(self, string: str) -> str:
        """Return space case string from a camelCase string."""
        return self._camel_pattern.sub(' ', string).lower()

    @cached_property
    def inflect(self) -> inflect.engine:
        """Return instance of inflect."""
        return inflect.engine()

    @staticmethod
    def random_string(string_length: int = 10) -> str:
        """Generate a random string of fixed length."""
        return ''.join(random.choice(ascii_letters) for _ in range(string_length))  # nosec

    @staticmethod
    def snake_string(string: str) -> 'SnakeString':
        """Return custom str with custom properties/methods."""
        return SnakeString(string)

    @staticmethod
    def snake_to_pascal(string: str) -> str:
        """Convert snake_case to PascalCase."""
        return string.replace('_', ' ').title().replace(' ', '')

    @staticmethod
    def snake_to_camel(string: str) -> str:
        """Convert snake_case to camelCase."""
        components = string.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    @staticmethod
    def to_bool(value: bool | int | str) -> bool:
        """Convert value to bool."""
        return str(value).lower() in ['1', 't', 'true', 'y', 'yes']

    @staticmethod
    def truncate_string(
        string: str,
        length: int,
        append_chars: str | None = '',
        spaces: bool = False,
    ) -> str:
        """Truncate a string to a given length.

        Args:
            string: The input string to truncate.
            length: The length of the truncated string.
            append_chars: Any character that should be appended to the
                string. Typically used for ellipsis (e.g. ...).
            spaces: If True truncation will be done at the
                nearest space before the truncation length to avoid chopping words.
        """
        if string is None:
            string = ''

        if length is None:
            length = len(string)

        if len(string) <= length:
            return string

        # set sane default for append_chars
        append_chars = str(append_chars or '')

        # ensure append_chars is not longer than length
        if len(append_chars) > length:  # pragma: no cover
            raise RuntimeError('Append chars cannot exceed the truncation length.')

        output = string[0 : length - len(append_chars)]
        if spaces is True:
            if not output.endswith(' '):
                # split output on spaces and drop last item to terminate string on word
                output = ' '.join(output.split(' ')[:-1])

        return f'{output.rstrip()}{append_chars}'

    @property
    def _camel_pattern(self) -> re.Pattern:
        """Return compiled re pattern for converting to camelCase."""
        return re.compile(r'(?<!^)(?=[A-Z])')

    @staticmethod
    def wrap_string(
        line: str,
        wrap_chars: list[str] | None = None,
        length: int = 100,
        force_wrap: bool = True,
    ) -> str:
        """Wrap a long string to a given length.

        Lines will only be broken on instances of strings from wrap_chars.

        Args:
            line: the string to break into lines
            wrap_chars: list of strings line should be broken on
            length: max length for any single line
            force_wrap: if True, line will be broken even if no string from wrap_chars is available.
        """
        wrap_chars = wrap_chars or [' ']

        # if line is already shorter than max length, we're all good!
        if len(line) < length:
            return line

        def _tokenize(acc, curr):
            """Tokenize the input into strings and separators (from wrap_chars)."""
            if curr in wrap_chars:
                return acc + [curr] + ['']
            if len(acc[-1]) == length and force_wrap:
                return acc + [curr]
            acc[-1] += curr
            return acc

        tokens = reduce(_tokenize, line, [''])

        def _chop(acc, curr):
            """Build strings in acc such that no string is greater than length chars.

            Breaks at chars in wrap_chars.
            """
            if len(acc[-1]) + len(curr) < length:
                acc[-1] += curr
                return acc
            return acc + [curr]

        lines = reduce(_chop, tokens, [''])

        return '\n'.join(lines).strip()


class CamelString(str):
    """Power String"""

    # properties
    so = StringOperation()

    def pascal_case(self):
        """Return a PascalCase version of a camelCase string."""
        return CamelString(self.so.camel_to_space(self).title().replace(' ', ''))

    def plural(self):
        """Return the plural spelling of a camelCase string."""
        return CamelString(self.so.inflect.plural(self.singular()))

    def singular(self):
        """Return the singular spelling of a camelCase string."""
        _singular = self.so.inflect.singular_noun(self)

        if not _singular:
            _singular = self
        return CamelString(_singular)

    def snake_case(self):
        """Return a snake_case version of a camelCase string."""
        return CamelString(self.so.camel_to_snake(self))

    def space_case(self):
        """Return a "space case" version of a camelCase string."""
        return CamelString(self.so.camel_to_space(self))


class SnakeString(str):
    """Power String"""

    # properties
    so = StringOperation()

    def camel_case(self):
        """Return a camelCase version of a snake_case string."""
        return SnakeString(self.so.snake_to_camel(self))

    def pascal_case(self):
        """Return a PascalCase version of a snake_case string."""
        return SnakeString(self.so.snake_to_pascal(self))

    def plural(self):
        """Return the plural spelling of a snake_case string."""
        return SnakeString(self.so.inflect.plural(self.singular()))

    def singular(self):
        """Return the singular spelling of a snake_case string."""
        _singular = self.so.inflect.singular_noun(self)

        if not _singular:
            _singular = self
        return SnakeString(_singular)

    def space_case(self, title: bool = True):
        """Return a "space case" version of a snake_case string."""
        space_string = self.replace('_', ' ')
        if title is True:
            space_string = space_string.title()
        return space_string



"""TcEx Framework Module"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class VictimAssetFilter(FilterABC):
    """Filter Object for VictimAssets"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.VICTIM_ASSETS.value

    def asset(self, operator: Enum, asset: list | str):
        """Filter Asset based on **asset** keyword.

        Args:
            operator: The operator enum for the filter.
            asset: The sub-type of the victim asset.
        """
        if isinstance(asset, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('asset', operator, asset, TqlType.STRING)

    def associated_group(self, operator: Enum, associated_group: int | list):
        """Filter associatedGroup based on **associatedGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_group: No description provided.
        """
        if isinstance(associated_group, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('associatedGroup', operator, associated_group, TqlType.INTEGER)

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    @property
    def has_victim_asset(self):
        """Return **VictimAssetFilter** for further filtering."""
        victim_assets = VictimAssetFilter(Tql())
        self._tql.add_filter('hasVictimAsset', TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)
        return victim_assets

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the victim asset.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the victim.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the victim.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The name of the victim asset.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def type(self, operator: Enum, type: int | list):  # pylint: disable=redefined-builtin
        """Filter Type ID based on **type** keyword.

        Args:
            operator: The operator enum for the filter.
            type: The ID of the victim asset type.
        """
        if isinstance(type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('type', operator, type, TqlType.INTEGER)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter Type Name based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The name of the victim asset type.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)

    def victim_id(self, operator: Enum, victim_id: int | list):
        """Filter Victim ID based on **victimId** keyword.

        Args:
            operator: The operator enum for the filter.
            victim_id: The ID of the victim the victim asset is applied to.
        """
        if isinstance(victim_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('victimId', operator, victim_id, TqlType.INTEGER)

    def victim_name(self, operator: Enum, victim_name: list | str):
        """Filter Victim Name based on **victimName** keyword.

        Args:
            operator: The operator enum for the filter.
            victim_name: The name of the victim.
        """
        if isinstance(victim_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('victimName', operator, victim_name, TqlType.STRING)



"""TcEx Framework Module"""
# standard library
import hmac
import time
from base64 import b64encode
from hashlib import sha256

# third-party
from requests import PreparedRequest, auth

from ...input.field_type.sensitive import Sensitive  # type: ignore # pylint: disable=import-error


class HmacAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, tc_api_access_id: str, tc_api_secret_key: Sensitive):
        """Initialize the Class properties."""
        auth.AuthBase.__init__(self)
        self.tc_api_access_id = tc_api_access_id
        self.tc_api_secret_key = tc_api_secret_key

    def _hmac_header(self, r: PreparedRequest, timestamp: float):
        """Return HMAC Authorization header value."""
        # define the signature using "full" path, HTTP method, and current timestamp
        signature = f'{r.path_url}:{r.method}:{timestamp}'

        # generate the sha256 signature using the tc secret key, encoded signature
        hmac_signature = hmac.new(
            self.tc_api_secret_key.value.encode(), signature.encode(), digestmod=sha256
        ).digest()

        # return the header value with access_id and b64 signature value
        return f'TC {self.tc_api_access_id}:{b64encode(hmac_signature).decode()}'

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """Add the authorization headers to the request."""
        timestamp = int(time.time())

        # Add required headers to auth.
        r.headers['Authorization'] = self._hmac_header(r, timestamp)
        r.headers['Timestamp'] = str(timestamp)
        return r



"""TcEx Framework Module"""
# standard library
from collections.abc import Generator

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidLengthValue, InvalidType


class Binary(bytes):
    """Binary Field Type"""

    allow_empty: bool = True
    max_length: int | None = None
    min_length: int | None = None
    strip: bool = False

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield cls.validate_type
        yield cls.validate_strip
        yield cls.validate_allow_empty
        yield cls.validate_max_length
        yield cls.validate_min_length

    @classmethod
    def validate_allow_empty(cls, value: bytes, field: ModelField) -> bytes:
        """Raise exception if value is empty and allow_empty is False."""
        if cls.allow_empty is False and value == b'':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @classmethod
    def validate_max_length(cls, value: bytes, field: ModelField) -> bytes:
        """Raise exception if value does not match pattern."""
        if cls.max_length is not None and len(value) > cls.max_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.max_length, operation='max'
            )
        return value

    @classmethod
    def validate_min_length(cls, value: bytes, field: ModelField) -> bytes:
        """Raise exception if value does not match pattern."""
        if cls.min_length is not None and len(value) < cls.min_length:
            raise InvalidLengthValue(
                field_name=field.name, constraint=cls.min_length, operation='min'
            )
        return value

    @classmethod
    def validate_strip(cls, value: bytes) -> bytes:
        """Raise exception if value is not a Binary type."""
        if value is not None and cls.strip is True:
            value = value.strip()
        return value

    @classmethod
    def validate_type(cls, value: bytes, field: ModelField) -> bytes:
        """Raise exception if value is not a Binary type."""
        if not isinstance(value, bytes):
            raise InvalidType(
                field_name=field.name, expected_types='(bytes)', provided_type=type(value)
            )
        return value


def binary(
    allow_empty: bool = True,
    min_length: int | None = None,
    max_length: int | None = None,
    strip: bool = False,
) -> Binary:
    """Return customized Binary type."""
    namespace = {
        'allow_empty': allow_empty,
        'max_length': max_length,
        'min_length': min_length,
        'strip': strip,
    }
    return type('CustomizedBinary', (Binary,), namespace)  # type: ignore



"""TcEx Framework Module"""

# first-party
from tcex.input.model.common_model import CommonModel
from tcex.input.model.playbook_common_model import PlaybookCommonModel
from tcex.input.model.service_model import ServiceModel


class AppApiServiceModel(CommonModel, PlaybookCommonModel, ServiceModel):
    """Model Definition"""



"""TcEx Framework Module"""
# pylint: disable=no-self-argument
# standard library
import re
from copy import deepcopy

# third-party
from pydantic import BaseModel, Field, root_validator, validator
from semantic_version import Version

from .install_json_model import (
    FeedsModel,
    InstallJsonCommonModel,
    InstallJsonOrganizationModel,
    OutputVariablesModel,
    ParamsModel,
    RetryModel,
    TypeEnum,
    snake_to_camel,
)
from .job_json_model import JobJsonCommonModel


class FeedsSpecModel(FeedsModel):
    """Model definition for app_spec.organization.feeds."""

    job: JobJsonCommonModel = Field(..., description='')


class NotesPerActionModel(BaseModel):
    """Model definition for app_spec.notes_per_action."""

    action: str = Field(..., description='The action name.')
    note: str = Field(..., description='The note describing the action.')

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class OrganizationModel(InstallJsonOrganizationModel):
    """Model definition for app_spec.organization."""

    feeds: list[FeedsSpecModel] = Field([], description='')

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class OutputVariablesSpecModel(OutputVariablesModel):
    """Model definition for app_spec.outputs.output_variables."""

    disabled: bool = Field(
        False,
        description='If True, the output will not be included in ij/lj files.',
    )
    type: str = Field(
        'String',
        description='The output variable type (e.g., String, TCEntity, etc).',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class OutputDataModel(BaseModel):
    """Model definition for app_spec.output_data."""

    display: str | None = Field(
        None,
        description='The display clause that controls visibility of the output.',
    )
    output_variables: list[OutputVariablesSpecModel] = Field(
        ...,
        description='An array of output variables.',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True

    @validator('display')
    def _display(cls, v: str):
        """Normalize "always True" expression for display clause."""
        if v is not None and v.lower() == '''tc_action not in ('')''':
            v = '1'
        return v  # pragma: no cover


class ParamsSpecModel(ParamsModel):
    """Model definition for app_spec.params."""

    display: str | None = Field(
        None,
        description='The display clause from the layout.json file.',
    )
    disabled: bool | None = Field(
        False,
        description='If True, the parameter will not be included in ij/lj files.',
    )
    type: TypeEnum = Field(
        TypeEnum.String,
        description='',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        fields = {'sequence': {'exclude': True}}
        smart_union = True
        use_enum_values = True
        validate_assignment = True


class PlaybookSpecModel(BaseModel):
    """Model definition for app_spec.playbook."""

    retry: RetryModel | None = Field(
        None,
        description='',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class ReleaseNoteModel(BaseModel):
    """Model definition for app_spec.releaseNotes."""

    notes: list[str] = Field(
        ...,
        description='One or more notes for the release.',
    )
    version: str = Field(
        ...,
        description='The version of the release.',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class SectionsModel(BaseModel):
    """Model definition for app_spec.sections."""

    section_name: str = Field(
        ...,
        description='The name of the section.',
    )
    params: list[ParamsSpecModel] = Field(
        ...,
        description='A list of input parameter data.',
    )

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True


class AppSpecYmlModel(InstallJsonCommonModel):
    """Model definition for the app_spec.yml file."""

    note_per_action: list[NotesPerActionModel] | None = Field(
        None,
        description='',
    )
    organization: OrganizationModel | None = Field(
        None,
        description='A section for settings related to the organization (job) Apps.',
    )
    output_data: list[OutputDataModel] | None = Field(
        None,
        description='The outputs data for Playbook and Service Apps.',
    )
    output_prefix: str | None = Field(
        None,
        description=(
            'The prefix for output variables, used for advanced request outputs. This value '
            'should match what is passed to the advanced request method in the playbook App.'
        ),
    )
    package_name: str = Field(
        None,
        description='The package name (app_name in tcex.json) for the App.',
    )
    playbook: PlaybookSpecModel | None = Field(
        None,
        description='The playbook section of the install.json.',
    )
    internal_notes: list[str] | None = Field(
        None,
        description='Internal notes for the App.',
    )
    release_notes: list[ReleaseNoteModel] = Field(
        ...,
        description='The release notes for the App.',
    )
    schema_version: str = Field(
        '1.0.0',
        description='The version of the App Spec schema.',
    )
    sections: list[SectionsModel] = Field(
        ...,
        description='Layout sections for an App including params.',
    )
    service_details: str | None = Field(
        None, description='Optional service details for Service Apps.'
    )

    @root_validator
    def _validate_no_input_duplication(cls, values):
        """Validate that no two parameters have the same name."""
        duplicates = {}
        for section in values.get('sections', []):
            for param in section.params:
                duplicates.setdefault(param.name, []).append(param.label)

        # strip out non-duplicates
        duplicates = {k: v for k, v in duplicates.items() if len(v) > 1}

        if duplicates:
            formatted_duplicates = [f'{k}({len(v)})' for k, v in duplicates.items()]
            raise ValueError(f'Found duplicate parameters: {", ".join(formatted_duplicates)}')
        return values

    @validator('schema_version')
    def _version(cls, v: str):
        """Return a version object for "version" fields."""
        if v is not None:
            return Version(v)
        return v  # pragma: no cover

    @validator('output_prefix', always=True, pre=True)
    def _output_prefix(cls, v: str | None, values: dict):
        """Validate output_prefix is set when required."""
        if 'advancedRequest' in values.get('features', []):
            if v is None:
                raise ValueError(
                    'The outputPrefix field is required when feature advancedRequest is enabled.'
                )
        else:
            # remove output_prefix if not required
            v = None
        return v

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        validate_assignment = True

    @property
    def inputs(self) -> list:
        """Return lj.inputs."""
        _inputs = []
        for sequence, section in enumerate(self.sections, start=1):
            # build params
            parameters = []
            for p in section.params:
                # exclude disabled and serviceConfig params
                if any([p.disabled, p.hidden, p.service_config]):
                    continue

                param = {'name': p.name}
                if p.display:
                    param['display'] = p.display

                parameters.append(param)

            if parameters:
                # append section
                _inputs.append(
                    {
                        'parameters': parameters,
                        'sequence': sequence,
                        'title': section.section_name,
                    }
                )
        return _inputs

    def get_note_per_action(self, action: str) -> NotesPerActionModel | None:
        """Return the note_per_action for the provided action."""
        for npa in self.note_per_action or []:
            if npa.action == action:
                return npa
        return None

    @property
    def note_per_action_formatted(self) -> list[str]:
        """Return formatted note_per_action."""
        _note_per_action = ['\n\nThe following actions are included:']
        _note_per_action.extend(
            [f'-   **{npa.action}** - {npa.note}' for npa in self.note_per_action or []]
        )
        return _note_per_action

    @property
    def outputs(self) -> list[OutputVariablesModel]:
        """Return lj.outputs."""
        _outputs = []
        for output_data in self.output_data or []:
            for output_variable in output_data.output_variables:
                if output_variable.disabled is True:
                    continue

                _outputs.append(
                    {
                        'display': output_data.display,
                        'name': output_variable.name,
                    }
                )
        return _outputs

    @property
    def output_variables(self) -> list[OutputVariablesModel]:
        """Return ij.playbook.outputVariables."""
        return [
            ov
            for output in self.output_data or []
            for ov in output.output_variables or []
            if ov.disabled is False
        ]

    @property
    def params(self) -> list[ParamsSpecModel]:
        """Return ij.params."""
        _params = []
        sequence = 1
        for section in deepcopy(self.sections):
            for param in section.params:
                if param.disabled is True:
                    continue

                # set default playbookDataType for String type params
                self._set_default_playbook_data_type(param)

                # set default validValues for String type params
                self._set_default_valid_values(param)

                # remove the disabled field (not supported in install.json)
                param.disabled = None

                # remove the display field (not supported in install.json)
                param.display = None

                # add the sequence number
                param.sequence = sequence

                _params.append(param)

                # increment sequence
                sequence += 1
        return _params

    @property
    def release_notes_formatted(self) -> list[str]:
        """Return readme_md.releaseNotes."""
        _release_notes = ['## Release Notes']
        _release_notes.append('')

        # try to sort release notes by version.  If we encounter any issues, just use release_notes
        try:
            sorted_releases = sorted(
                (r for r in self.release_notes),
                key=lambda r: Version(re.match(r'^\d+.\d+.\d+', r.version)[0]),  # type: ignore
                reverse=True,
            )
        except Exception:
            sorted_releases = self.release_notes

        for release_note in sorted_releases:
            _release_notes.append(f'### {release_note.version}')
            _release_notes.append('')
            _release_notes.extend([f'-   {rn}' for rn in release_note.notes])
            _release_notes.append('')
        return _release_notes

    @property
    def requires_layout(self):
        """Return True if App requires a layout.json file."""
        if self.runtime_level.lower() in ['apiservice', 'feedapiservice', 'organization']:
            return False

        for section in self.sections:
            for param in section.params:
                if param.display:
                    return True

        for output_data in self.output_data or []:
            if output_data.display not in [None, '1', '']:
                return True

        return False

    def _set_default_playbook_data_type(self, param: ParamsSpecModel):
        """Set default playbookDataType for String type params.

        Playbook Data Types rule:
          * Input type is "String"
          * No "playbookDataType" values are provided
        """
        if self.runtime_level.lower() == 'playbook':
            # by rule any input of type String must have String and playbookDataType
            if (
                param.type in ['EditChoice', 'KeyValueList', 'String']
                and not param.playbook_data_type
            ):
                param.playbook_data_type = ['String']

    def _set_default_valid_values(self, param: ParamsSpecModel):
        """Set default playbookDataType for String type params.

        Valid Values rule:
          * Input type is "String"
          * No "validValues" values are provided
          * The playbookDataType supports String
        """
        if (
            param.type in ['KeyValueList', 'String']
            and not param.valid_values
            and (
                'String' in param.playbook_data_type
                or self.runtime_level.lower()
                in ['organization', 'triggerservice', 'webhooktriggerservice']
            )
        ):
            param.valid_values = ['${TEXT}']
            if param.encrypt is True:
                param.valid_values = ['${KEYCHAIN}']



"""TcEx Framework Module"""
# standard library
import ast
import logging
import os
import re

# third-party
import black
import isort
from black.report import NothingChanged

from .render.render import Render

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class CodeOperation:
    """TcEx FrameWork Code Operations"""

    @staticmethod
    def find_line_in_code(
        needle: str | re.Pattern,
        code: str,
        trigger_start: re.Pattern | str | None = None,
        trigger_stop: re.Pattern | str | None = None,
    ) -> str | None:
        """Return matching line of code in a class definition.

        Args:
            needle: The string to search for.
            code: The contents of the Python file to search.
            trigger_start: The regex pattern to use to trigger the search.
            trigger_stop: The regex pattern to use to stop the search.
        """
        magnet_on = not trigger_start
        for line in ast.unparse(ast.parse(code)).split('\n'):
            if line.lstrip()[:1] not in ("'", '"'):
                # Find class before looking for needle
                if trigger_start is not None and re.match(trigger_start, line):
                    magnet_on = True
                    continue

                # find need now that class definition is found
                if magnet_on is True and re.match(needle, line):
                    line = line.strip()
                    return line

                # break if needle not found before next class definition
                if trigger_stop is not None and re.match(trigger_stop, line) and magnet_on is True:
                    break
        return None

    @staticmethod
    def find_line_number(
        needle: str,
        contents: str,
        trigger_start: re.Pattern | str | None = None,
        trigger_stop: re.Pattern | str | None = None,
    ) -> int | None:
        """Return matching line of code in a class definition.

        Args:
            needle: The string to search for.
            contents: The contents (haystack) to search
            trigger_start: The regex pattern to use to trigger the search.
            trigger_stop: The regex pattern to use to stop the search.
        """
        magnet_on = not trigger_start
        for line_number, line in enumerate(contents.split('\n'), start=1):
            if line.strip():
                # set magnet_on to True if trigger_start is found
                if trigger_start is not None and re.match(trigger_start, line):
                    magnet_on = True
                    continue

                # find needle now that trigger is found
                if magnet_on is True and re.match(needle, line):
                    return line_number

                # break if trigger_stop is defined and found
                if trigger_stop is not None and re.match(trigger_stop, line) and magnet_on is True:
                    break
        return None

    @staticmethod
    def format_code(_code):
        """Return formatted code."""
        # run black formatter on code
        mode = black.FileMode(line_length=100, string_normalization=False)
        try:
            _code = black.format_file_contents(_code, fast=False, mode=mode)
        except ValueError as ex:
            _logger.exception('Formatting of code with black failed.')
            Render.panel.failure(f'Formatting of code with black failed {ex}.')
        except NothingChanged:
            pass

        # run isort on code
        try:
            isort_args = (
                {'settings_file': 'pyproject.toml'} if os.path.isfile('pyproject.toml') else {}
            )
            isort_config = isort.Config(**isort_args)  # type: ignore
            _code = isort.code(_code, config=isort_config)
        except Exception as ex:
            _logger.exception('Formatting of code with isort failed.')
            # _logger.debug(_code)
            Render.panel.failure(f'Formatting of code with isort failed {ex}.')

        return _code



"""TcEx Framework Module"""
# standard library
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .install_json import InstallJson


class InstallJsonValidate:
    """Config object for install.json file (validator)"""

    def __init__(self, ij: 'InstallJson'):  # pylint: disable=E0601
        """Initialize instance properties."""
        self.ij = ij

    def validate_duplicate_input(self) -> list:
        """Check for duplicate input names."""
        duplicates = []
        tracker = []
        for param in self.ij.model.params:
            if param.name in tracker:
                duplicates.append(param.name)
            tracker.append(param.name)
        return duplicates

    def validate_duplicate_output(self) -> list:
        """Check for duplicate input names."""
        duplicates = []
        tracker = []
        if self.ij.model.playbook is None:
            return duplicates

        for output in self.ij.model.playbook.output_variables or []:
            name_type = f'{output.name}-{output.type}'
            if name_type in tracker:
                duplicates.append(output.name)
            tracker.append(name_type)
        return duplicates

    def validate_duplicate_sequence(self) -> list:
        """Check for duplicate sequence numbers."""
        duplicates = []
        tracker = []
        for param in self.ij.model.params:
            if param.sequence in tracker:
                duplicates.append(param.sequence)
            tracker.append(param.sequence)
        return duplicates



"""TcEx Framework Module"""
# standard library
import collections
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime
from inspect import signature
from typing import Any

# third-party
import jmespath
from jmespath import functions

# first-party
from tcex.api.tc.ti_transform.model import AttributeTransformModel  # TYPE-CHECKING
from tcex.api.tc.ti_transform.model import SecurityLabelTransformModel  # TYPE-CHECKING
from tcex.api.tc.ti_transform.model import TagTransformModel  # TYPE-CHECKING
from tcex.api.tc.ti_transform.model import (
    GroupTransformModel,
    IndicatorTransformModel,
    MetadataTransformModel,
)
from tcex.api.tc.ti_transform.model.transform_model import (
    AssociatedGroupTransform,
    DatetimeTransformModel,
    FileOccurrenceTransformModel,
)
from tcex.logger.trace_logger import TraceLogger
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class TcFunctions(functions.Functions):
    """ThreatConnect custom jmespath functions."""

    @functions.signature({'types': ['array']}, {'types': ['string']})
    def _func_null_leaf(self, arr: list, search: str) -> list:
        """Return value in array even if they are null."""
        return [a.get(search) for a in arr]

    @functions.signature({'types': ['array']}, {'types': ['string']})
    def _func_delete(self, arr: list, search: str) -> list:
        """Return array after popping value at address out."""
        for a in arr:
            try:
                a.pop(search)
            except Exception:
                _logger.debug('Skipping delete, field not found.')
        return arr


class TransformsABC(ABC):
    """Transform Abstract Base Class"""

    def __init__(
        self,
        ti_dicts: list[dict],
        transforms: list[GroupTransformModel | IndicatorTransformModel],
    ):
        """Initialize instance properties."""
        self.ti_dicts = ti_dicts
        self.transforms = transforms

        # properties
        self.log = _logger
        self.transformed_collection: list[TransformABC] = []

        # validate transforms
        self._validate_transforms()

    def _validate_transforms(self):
        """Validate the transform model."""
        if len(self.transforms) > 1:
            for transform in self.transforms:
                if transform.applies is None:
                    raise ValueError(
                        'If more than one transform is provided, each '
                        'provided transform must provide an apply field.',
                    )


class TransformABC(ABC):
    """Transform Abstract Base Class"""

    def __init__(
        self,
        ti_dict: dict,
        transforms: list[GroupTransformModel | IndicatorTransformModel],
    ):
        """Initialize instance properties."""
        self.ti_dict = ti_dict
        self.transforms = transforms if isinstance(transforms, list) else [transforms]

        # properties
        self.adhoc_groups: list[dict] = []
        self.adhoc_indicators: list[dict] = []
        self.log = _logger
        # the current active transform
        self.transform: GroupTransformModel | IndicatorTransformModel
        self.transformed_item = {}
        self.util = Util()
        self.jmespath_options = jmespath.Options(
            custom_functions=TcFunctions(), dict_cls=collections.OrderedDict
        )

        # validate transforms
        self._validate_transforms()

    @staticmethod
    def _always_array(value: str | list | None) -> list:
        """Ensure value is always an array."""
        if value is None:
            value = []
        elif not isinstance(value, list):
            value = [value]
        return value

    @staticmethod
    def _build_summary(
        val1: str | None = None, val2: str | None = None, val3: str | None = None
    ) -> str:
        """Build the Indicator summary using available values."""
        return ' : '.join([value for value in [val1, val2, val3] if value is not None])

    def _path_search(self, path: str) -> Any:
        """Return the value of the provided path.

        Path can return any type of data from the TI dict.
        """
        if path is not None:
            value = jmespath.search(path, self.ti_dict, options=self.jmespath_options)
            # self.log.trace(f'feature=transform, action=path-search, path={path}, value={value}')
            return value
        return None

    def _process(self):
        """Process the TI data."""
        # choose the correct transform model before processing TI data
        self._select_transform()

        # process type first, fail early
        self._process_type()

        # process type specific data
        if isinstance(self.transform, GroupTransformModel):
            self._process_group()
        elif isinstance(self.transform, IndicatorTransformModel):
            self._process_indicator()

        # self.process_associations(self.transform.associations)
        self._process_associated_group(self.transform.associated_groups)
        self._process_attributes(self.transform.attributes or [])
        self._process_security_labels(self.transform.security_labels or [])
        self._process_tags(self.transform.tags or [])

        # date added
        self._process_metadata_datetime('dateAdded', self.transform.date_added)

        # last modified
        self._process_metadata_datetime('lastModified', self.transform.last_modified)

        # xid
        self._process_metadata('xid', self.transform.xid)

    def _process_associated_group(self, associations: list[AssociatedGroupTransform]):
        """Process Attribute data"""
        for association in associations or []:
            for value in filter(bool, self._process_metadata_transform_model(association.value)):
                self.add_associated_group(value)  # type: ignore

    def _process_metadata_transform_model(
        self, value: bool | MetadataTransformModel | str | None, expected_length: int | None = None
    ) -> list:
        """Process fields that can be static values or a MetadataTransformModel.

        If value is not a MetadataTransformModel (i.e., it's a static value), and expected_length
        is given, "spread" the static value into an array of expected_length length.

        """
        if value is None:
            if expected_length:
                return [value] * expected_length
            return []

        if isinstance(value, MetadataTransformModel):
            transformed_value = self._transform_values(value)

            if expected_length is not None and len(transformed_value) != expected_length:
                raise RuntimeError(
                    f'Expected transform value of length {expected_length} for {value}, '
                    f'but length was {len(transformed_value)}'
                )

            return transformed_value

        if expected_length is not None:
            transformed_value = [value] * expected_length
        else:
            transformed_value = [value]

        return transformed_value

    def _process_attributes(self, attributes: list[AttributeTransformModel]):
        """Process Attribute data"""
        for attribute in attributes or []:
            values = self._process_metadata_transform_model(attribute.value)
            if not values:
                # self.log.info(f'No values found for attribute transform {attribute.dict()}')
                continue

            types = self._process_metadata_transform_model(attribute.type, len(values))
            source = self._process_metadata_transform_model(attribute.source, len(values))
            displayed = self._process_metadata_transform_model(attribute.displayed, len(values))

            param_keys = ['type_', 'value', 'displayed', 'source']
            params = [dict(zip(param_keys, p)) for p in zip(types, values, displayed, source)]

            for param in params:
                param = self.util.remove_none(param)
                if 'value' not in param:
                    continue

                if 'type_' not in param:
                    self.log.warning(
                        'feature=transform, action=process-attribute, '
                        f'transform={attribute.dict(exclude_unset=True)}, error=no-type'
                    )
                    continue

                # strip out None params so that required params are enforced and optional
                # params with default values are respected.
                try:
                    self.add_attribute(**param)
                except Exception:
                    self.log.exception(
                        'feature=transform, action=process-attribute, '
                        f'transform={attribute.dict(exclude_unset=True)}'
                    )

    def _process_file_occurrences(self, file_occurrences: list[FileOccurrenceTransformModel]):
        """Process File Occurrences data.

        File Occurrences are a bit weird, in that none of the fields are required.  Because of this,
        There may be results where all of the fields are None, in which case we'll skip that result
        and not call add_file_occurrence.
        """
        for file_occurrence in file_occurrences or []:
            expected_length = max(
                map(
                    lambda f: len(self._process_metadata_transform_model(f)),
                    [
                        field
                        for field in (
                            file_occurrence.file_name,
                            file_occurrence.path,
                            file_occurrence.date,
                        )
                        if field is not None
                    ],
                )
            )
            file_name = self._process_metadata_transform_model(
                file_occurrence.file_name, expected_length
            )
            path = self._process_metadata_transform_model(file_occurrence.path, expected_length)
            date = self._process_metadata_transform_model(file_occurrence.date, expected_length)

            param_keys = ['file_name', 'path', 'date']
            params = [dict(zip(param_keys, p)) for p in zip(file_name, path, date)]

            for kwargs in filter(bool, params):  # get rid of empty dicts
                self.add_file_occurrence(**self.util.remove_none(kwargs))

    def _process_confidence(self, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        self.add_confidence(self._transform_value(metadata))

    def _process_group(self):
        """Process Group Specific data."""
        if not isinstance(self.transform, GroupTransformModel):
            return

        self._process_name()

        if self.transformed_item['type'] == 'Campaign':
            self._process_metadata_datetime('firstSeen', self.transform.first_seen)

        if self.transformed_item['type'] == 'Document':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata('malware', self.transform.malware)
            self._process_metadata('password', self.transform.password)

        if self.transformed_item['type'] == 'Email':
            self._process_metadata('body', self.transform.body)
            self._process_metadata('from', self.transform.from_addr)
            self._process_metadata('header', self.transform.header)
            self._process_metadata('subject', self.transform.subject)
            self._process_metadata('to', self.transform.to_addr)

        if self.transformed_item['type'] in ('Event', 'Incident'):
            self._process_metadata_datetime('eventDate', self.transform.event_date)
            self._process_metadata('status', self.transform.status)

        if self.transformed_item['type'] == 'Report':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata_datetime('publishDate', self.transform.publish_date)

        # Handle sig specific fields here
        if self.transformed_item['type'] == 'Signature':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata('fileType', self.transform.file_type)
            self._process_metadata('fileText', self.transform.file_text)

    def _process_indicator(self):
        """Process Indicator Specific data."""
        if not isinstance(self.transform, IndicatorTransformModel):
            return

        # handle the 3 possible indicator fields
        self._process_indicator_values()

        if self.transform.active:
            self._process_metadata('active', self.transform.active)

        self._process_confidence(self.transform.confidence)
        self._process_rating(self.transform.rating)

        if self.transformed_item['type'] == 'File':
            self._process_metadata('size', self.transform.size)
            self._process_file_occurrences(self.transform.file_occurrences or [])

        if self.transformed_item['type'] == 'Host':
            self._process_metadata('dnsActive', self.transform.dns_active)
            self._process_metadata('whoisActive', self.transform.whois_active)

    def _process_indicator_values(self):
        """Process Indicator value."""
        if not isinstance(self.transform, IndicatorTransformModel):
            return

        value1 = self._transform_value(self.transform.value1)
        value2 = self._transform_value(self.transform.value2)
        value3 = self._transform_value(self.transform.value3)

        if not any([value1, value2, value3]):
            self.log.error(
                'feature=ti-transform, event=process-indicators, message=no-indicator-value-found, '
                f'path-value1={value1}, path-value2={value2}, path-value3={value3}'
            )
            raise RuntimeError('At least one indicator value must be provided.')

        self.add_summary(self._build_summary(value1, value2, value3))

    def _process_name(self):
        """Process Group Name data."""
        if not isinstance(self.transform, GroupTransformModel):
            return

        name = self._transform_value(self.transform.name)

        if name is None:
            self.log.error(
                'feature=ti-transform, event=process-group-name, message=no-name-found, '
                f'path={self.transform.name.path}'
            )
            raise RuntimeError('At least one indicator value must be provided.')

        self.add_name(name)

    def _process_metadata(self, key: str, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        value = self._transform_value(metadata)
        if value is not None:
            self.add_metadata(key, value)

    def _process_metadata_datetime(self, key: str, metadata: DatetimeTransformModel | None):
        """Process metadata fields that should be a TC datetime."""
        if metadata is not None and metadata.path is not None:
            value = self._path_search(metadata.path)
            if value is not None:
                self.add_metadata(
                    key, self.util.any_to_datetime(value).strftime('%Y-%m-%dT%H:%M:%SZ')
                )

    def _process_security_labels(self, labels: list[SecurityLabelTransformModel]):
        """Process Tag data"""
        for label in labels or []:
            names = self._process_metadata_transform_model(label.value)
            if not names:
                # self.log.info(f'No values found for security label transform {label.dict()}')
                continue

            descriptions = self._process_metadata_transform_model(
                label.description, expected_length=len(names)
            )
            colors = self._process_metadata_transform_model(label.color, expected_length=len(names))

            param_keys = ['color', 'description', 'name']
            params = [dict(zip(param_keys, p)) for p in zip(colors, descriptions, names)]

            for kwargs in params:
                kwargs = self.util.remove_none(kwargs)
                if 'name' not in kwargs:
                    continue
                # strip out None params so that required params are enforced and optional
                # params with default values are respected.
                self.add_security_label(**kwargs)

    def _process_tags(self, tags: list[TagTransformModel]):
        """Process Tag data"""
        for tag in tags or []:
            for value in filter(bool, self._process_metadata_transform_model(tag.value)):
                self.add_tag(name=value)  # type: ignore

    def _process_rating(self, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        self.add_rating(self._transform_value(metadata))

    def _process_type(self):
        """Process standard metadata fields."""
        _type = self._transform_value(self.transform.type)
        if _type is not None:
            self.transformed_item['type'] = _type
        else:
            self.log.error(
                'feature=ti-transform, action=process-type, error=invalid=type, '
                f'path={self.transform.type.path}, value={_type}'
            )
            raise RuntimeError('Invalid type')

    def _select_transform(self):
        """Select the correct transform based on the "applies" field."""
        for transform in self.transforms:
            if transform.applies is None or transform.applies(self.ti_dict) is True:
                self.transform = transform
                break
        else:
            raise RuntimeError('No transform found for TI data')

    def _transform_value(self, metadata: MetadataTransformModel | None) -> str | None:
        """Pass value to series transforms."""
        # not all fields are required
        if metadata is None:
            return None

        # not all metadata fields have a path, but they must have a path or default
        if metadata.path is None:
            return metadata.default

        # get value from path
        value = self._path_search(metadata.path)

        # return default if value of None is returned from Path
        # IMPORTANT: a value of None passed to the transform may cause a failure (lambda x.lower())
        if value is None:
            return metadata.default

        for t in metadata.transform or []:
            if isinstance(value, str):
                # pass value to static_map or callable, but never both
                if t.filter_map is not None:
                    value = self._transform_value_map(value, t.filter_map, True)
                elif t.static_map is not None:
                    value = self._transform_value_map(value, t.static_map)
                elif callable(t.method):
                    value = self._transform_value_callable(value, t.method, t.kwargs)

        # ensure only a string value or None is returned (set to default if required)
        if value is None:
            value = metadata.default
        elif not isinstance(value, str):
            value = str(value)

        return value

    def _transform_value_callable(
        self, value: dict | list | str, c: Callable, kwargs=None
    ) -> str | None | list[str] | None:
        """Transform values in the TI data."""
        # find signature of method and call with correct args
        kwargs = kwargs or {}
        try:
            sig = signature(c, follow_wrapped=True)
            if 'ti_dict' in sig.parameters:
                kwargs.update({'ti_dict': self.ti_dict})
            if 'transform' in sig.parameters:
                kwargs.update({'transform': self})
        except ValueError:  # signature doesn't work for many built-in methods/functions
            pass

        # pass value to transform callable/method, which should always return a string
        return c(value, **kwargs)

    def _transform_value_map(self, value: str, map_: dict, passthrough: bool = False) -> str:
        """Transform a value using a static map."""
        _default = value if passthrough is True else None
        if isinstance(value, str):
            # a static map is a dict of key/value pairs
            value = map_.get(value.lower(), _default)
        else:
            self.log.warning(
                f'''feature=ti-transform, action=transform-value, '''
                f'''message='static-map-requires-str-value", value={value}'''
            )
        return value

    def _transform_values(self, metadata: MetadataTransformModel | None) -> list[str]:
        """Pass value to series transforms."""

        def _default() -> list:
            """Return default value (as list) if exists, else empty list."""
            if metadata is None or metadata.default is None:
                return []

            return self._always_array(metadata.default)

        # not all items have all metadata fields
        if metadata is None:
            # self.log.trace('feature=transform, action=transform-values, metadata=None')
            return []

        # not all metadata fields have a path, but they must have a path or default
        if metadata.path is None:
            default = _default()
            # self.log.trace(
            #     f'feature=transform, action=transform-values, metadata-path=None, value={default}'
            # )
            return default

        # path search can return multiple data types and single or multiple values
        value = self._path_search(metadata.path)

        # return default if value of None is returned from Path
        # IMPORTANT: a None value passed to the transform may cause a failure (lambda x: x.lower())
        if value in [None, []]:
            default = _default()
            # self.log.trace(
            #     f'feature=transform, action=transform-values, metadata-path=None, value={default}'
            # )
            return default

        for t in metadata.transform or []:
            if t.filter_map is not None:
                # when path search returns an array of values, each value is mapped
                value = [
                    self._transform_value_map(v, t.filter_map, True)
                    for v in self._always_array(value)
                ]
            elif t.static_map is not None:
                # when path search returns an array of values, each value is mapped
                _values = []
                for v in self._always_array(value):
                    v = self._transform_value_map(v, t.static_map)
                    if v is not None:
                        _values.append(v)
                value = _values
            # PYRIGHT-MISS - None check for value already performed above
            elif callable(t.method) and value is not None:
                value = self._transform_value_callable(value, t.method, t.kwargs)
            elif callable(t.for_each):
                value = [
                    self._transform_value_callable(v, t.for_each, t.kwargs) if v is not None else v
                    for v in self._always_array(value)
                ]

        # the output should be an array of strings or empty array
        _value = []
        for v in self._always_array(value):
            if v in [None, '']:
                if metadata.default is not None:
                    _value.append(metadata.default)
                else:
                    _value.append(v)
            else:
                _value.append(v)

        # self.log.trace(f'feature=transform, action=transform-values, value={_value}')
        return _value

    def _validate_transforms(self):
        """Validate the transform model."""
        if len(self.transforms) > 1:
            for transform in self.transforms:
                if transform.applies is None:
                    raise ValueError(
                        'If more than one transform is provided, each '
                        'provided transform must provide an apply field.',
                    )

    @abstractmethod
    def add_associated_group(self, group_xid: str):
        """Abstract method"""

    @abstractmethod
    def add_attribute(
        self,
        type_: str,
        value: str,
        displayed: bool = False,
        source: str | None = None,
    ):
        """Abstract method"""

    @abstractmethod
    def add_file_occurrence(
        self,
        file_name: str | None = None,
        path: str | None = None,
        date: datetime | None = None,
    ):
        """Abstract method"""

    @abstractmethod
    def add_confidence(self, confidence: int | str | None):
        """Abstract method"""

    @abstractmethod
    def add_metadata(self, key: str, value: str):
        """Abstract method"""

    @abstractmethod
    def add_name(self, name: str | None):
        """Abstract method"""

    @abstractmethod
    def add_rating(self, rating: float | int | str | None):
        """Abstract method"""

    @abstractmethod
    def add_security_label(
        self, name: str, color: str | None = None, description: str | None = None
    ):
        """Abstract method"""

    @abstractmethod
    def add_summary(self, value: str | None):
        """Abstract method"""

    @abstractmethod
    def add_tag(self, name: str):
        """Abstract method"""



"""TcEx Framework Module"""
# standard library
import json
import logging
from collections import OrderedDict
from functools import cached_property
from pathlib import Path

from ...pleb.singleton import Singleton
from .model.install_json_model import OutputVariablesModel, ParamsModel
from .model.layout_json_model import LayoutJsonModel

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class LayoutJson(metaclass=Singleton):
    """Config object for layout.json configuration file"""

    def __init__(
        self,
        filename: str | None = None,
        path: Path | str | None = None,
        logger: logging.Logger | None = None,
    ):
        """Initialize instance properties."""
        filename = filename or 'layout.json'
        path = Path(path or Path.cwd())
        self.log = logger or _logger

        # properties
        self.fqfn = path / filename

    @cached_property
    def contents(self) -> dict:
        """Return layout.json file contents."""
        contents = {}
        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    contents = json.load(fh, object_pairs_hook=OrderedDict)
            except (OSError, ValueError):  # pragma: no cover
                self.log.error(
                    f'feature=layout-json, exception=failed-reading-file, filename={self.fqfn}'
                )
        else:  # pragma: no cover
            self.log.error(f'feature=layout-json, exception=file-not-found, filename={self.fqfn}')
        return contents

    def create(self, inputs: list[ParamsModel], outputs: list[OutputVariablesModel]):
        """Create new layout.json file based on inputs and outputs."""

        def input_data(sequence: int, title: str) -> dict:
            return {
                'parameters': [],
                'sequence': sequence,
                'title': title,
            }

        lj = LayoutJsonModel(
            **{
                'inputs': [
                    input_data(1, 'Action'),
                    input_data(2, 'Connection'),
                    input_data(3, 'Configure'),
                    input_data(4, 'Advanced'),
                ],
                'outputs': [{'display': '', 'name': o.name} for o in outputs],
            }
        )

        for input_ in inputs:
            if input_.name == 'tc_action':
                lj.inputs[0].parameters.append({'name': 'tc_action'})  # type: ignore
            elif input_.hidden is True:
                lj.inputs[2].parameters.append(
                    {
                        'display': "'hidden' != 'hidden'",
                        'hidden': 'true',
                        'name': input_.name,
                    }  # type: ignore
                )
            else:
                lj.inputs[2].parameters.append({'display': '', 'name': input_.name})  # type: ignore

        # write layout file to disk
        data = lj.json(
            by_alias=True, exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True
        )
        self.write(data)

    @property
    def has_layout(self):
        """Return True if App has layout.json file."""
        return self.fqfn.is_file()

    @cached_property
    def model(self) -> LayoutJsonModel:
        """Return the Install JSON model."""
        return LayoutJsonModel(**self.contents)

    @property
    def update(self):
        """Return InstallJsonUpdate instance."""
        return LayoutJsonUpdate(lj=self)

    def write(self, data: str):
        """Write updated file.

        Args:
            data: The JSON string to write data.
        """
        with self.fqfn.open(mode='w') as fh:
            fh.write(f'{data}\n')


class LayoutJsonUpdate:
    """Update layout.json file with current standards and schema."""

    def __init__(self, lj: LayoutJson):
        """Initialize instance properties."""
        self.lj = lj

    def multiple(self):
        """Update the layouts.json file."""
        # APP-86 - sort output data by name
        self.update_sort_outputs()

        data = self.lj.model.json(
            by_alias=True, exclude_defaults=True, exclude_none=True, indent=2, sort_keys=True
        )
        self.lj.write(data)

    def update_sort_outputs(self):
        """Sort output field by name."""
        # APP-86 - sort output data by name
        self.lj.model.outputs = sorted(
            self.lj.model.dict().get('outputs', []), key=lambda i: i['name']  # type: ignore
        )
